from typing import Any, Dict

import pytest
from hypothesis import assume, given, note, settings, strategies

import xgboost as xgb
from xgboost import testing as tm

pytestmark = tm.timeout(10)

parameter_strategy = strategies.fixed_dictionaries(
    {
        "booster": strategies.just("gblinear"),
        "eta": strategies.floats(0.01, 0.25),
        "tolerance": strategies.floats(1e-5, 1e-2),
        "nthread": strategies.integers(1, 4),
        "feature_selector": strategies.sampled_from(
            ["cyclic", "shuffle", "greedy", "thrifty"]
        ),
        "top_k": strategies.integers(1, 10),
    }
)


def train_result(
    param: Dict[str, Any], dmat: xgb.DMatrix, num_rounds: int
) -> Dict[str, Any]:
    result: Dict[str, Any] = {}
    booster = xgb.train(
        param,
        dmat,
        num_rounds,
        [(dmat, "train")],
        verbose_eval=False,
        evals_result=result,
    )
    assert booster.num_boosted_rounds() == num_rounds
    return result


class TestGPULinear:
    @given(parameter_strategy, strategies.integers(10, 50), tm.make_dataset_strategy())
    @settings(deadline=None, max_examples=20, print_blob=True)
    def test_gpu_coordinate(
        self, param: Dict[str, Any], num_rounds: int, dataset: tm.TestDataset
    ) -> None:
        assume(len(dataset.y) > 0)
        param["updater"] = "coord_descent"
        param["device"] = "cuda"
        param = dataset.set_params(param)
        result = train_result(param, dataset.get_dmat(), num_rounds)["train"][
            dataset.metric
        ]
        note(result)
        assert tm.non_increasing(result)

    # Loss is not guaranteed to always decrease because of regularisation parameters
    # We test a weaker condition that the loss has not increased between the first and last
    # iteration
    @given(
        parameter_strategy,
        strategies.integers(10, 50),
        tm.make_dataset_strategy(),
        strategies.floats(1e-5, 0.8),
        strategies.floats(1e-5, 0.8),
    )
    @settings(deadline=None, max_examples=20, print_blob=True)
    def test_gpu_coordinate_regularised(
        self,
        param: Dict[str, Any],
        num_rounds: int,
        dataset: tm.TestDataset,
        alpha: float,
        lambd: float,
    ) -> None:
        assume(len(dataset.y) > 0)
        param["updater"] = "coord_descent"
        param["device"] = "cuda"
        param["alpha"] = alpha
        param["lambda"] = lambd
        param = dataset.set_params(param)
        result = train_result(param, dataset.get_dmat(), num_rounds)["train"][
            dataset.metric
        ]
        note(result)
        assert tm.non_increasing([result[0], result[-1]])

    @pytest.mark.skipif(**tm.no_cupy())
    def test_gpu_coordinate_from_cupy(self) -> None:
        # Training linear model is quite expensive, so we don't include it in
        # test_from_cupy.py
        import cupy

        params = {
            "booster": "gblinear",
            "updater": "coord_descent",
            "device": "cuda",
            "n_estimators": 100,
        }
        X, y = tm.get_california_housing()
        cpu_model = xgb.XGBRegressor(**params)
        cpu_model.fit(X, y)
        cpu_predt = cpu_model.predict(X)

        X = cupy.array(X)
        y = cupy.array(y)
        gpu_model = xgb.XGBRegressor(**params)
        gpu_model.fit(X, y)
        gpu_predt = gpu_model.predict(X)
        cupy.testing.assert_allclose(cpu_predt, gpu_predt)



"""
Collection of examples for using xgboost.spark estimator interface
==================================================================

@author: Weichen Xu
"""

import numpy as np
import sklearn.datasets
from pyspark.ml.evaluation import MulticlassClassificationEvaluator, RegressionEvaluator
from pyspark.ml.linalg import Vectors
from pyspark.sql import DataFrame, SparkSession
from pyspark.sql.functions import rand
from sklearn.model_selection import train_test_split

from xgboost.spark import SparkXGBClassifier, SparkXGBRegressor

spark = SparkSession.builder.master("local[*]").getOrCreate()


def create_spark_df(X: np.ndarray, y: np.ndarray) -> DataFrame:
    return spark.createDataFrame(
        spark.sparkContext.parallelize(
            [(Vectors.dense(features), float(label)) for features, label in zip(X, y)]
        ),
        ["features", "label"],
    )


# load diabetes dataset (regression dataset)
diabetes_X, diabetes_y = sklearn.datasets.load_diabetes(return_X_y=True)
diabetes_X_train, diabetes_X_test, diabetes_y_train, diabetes_y_test = train_test_split(
    diabetes_X, diabetes_y, test_size=0.3, shuffle=True
)

diabetes_train_spark_df = create_spark_df(diabetes_X_train, diabetes_y_train)
diabetes_test_spark_df = create_spark_df(diabetes_X_test, diabetes_y_test)

# train xgboost regressor model
xgb_regressor = SparkXGBRegressor(max_depth=5)
xgb_regressor_model = xgb_regressor.fit(diabetes_train_spark_df)

transformed_diabetes_test_spark_df = xgb_regressor_model.transform(
    diabetes_test_spark_df
)
regressor_evaluator = RegressionEvaluator(metricName="rmse")
print(
    f"regressor rmse={regressor_evaluator.evaluate(transformed_diabetes_test_spark_df)}"
)

diabetes_train_spark_df2 = diabetes_train_spark_df.withColumn(
    "validationIndicatorCol", rand(1) > 0.7
)

# train xgboost regressor model with validation dataset
xgb_regressor2 = SparkXGBRegressor(
    max_depth=5, validation_indicator_col="validationIndicatorCol"
)
xgb_regressor_model2 = xgb_regressor2.fit(diabetes_train_spark_df2)
transformed_diabetes_test_spark_df2 = xgb_regressor_model2.transform(
    diabetes_test_spark_df
)
print(
    f"regressor2 rmse={regressor_evaluator.evaluate(transformed_diabetes_test_spark_df2)}"
)


# load iris dataset (classification dataset)
iris_X, iris_y = sklearn.datasets.load_iris(return_X_y=True)
iris_X_train, iris_X_test, iris_y_train, iris_y_test = train_test_split(
    iris_X, iris_y, test_size=0.3, shuffle=True
)

iris_train_spark_df = create_spark_df(iris_X_train, iris_y_train)
iris_test_spark_df = create_spark_df(iris_X_test, iris_y_test)

# train xgboost classifier model
xgb_classifier = SparkXGBClassifier(max_depth=5)
xgb_classifier_model = xgb_classifier.fit(iris_train_spark_df)

transformed_iris_test_spark_df = xgb_classifier_model.transform(iris_test_spark_df)
classifier_evaluator = MulticlassClassificationEvaluator(metricName="f1")
print(f"classifier f1={classifier_evaluator.evaluate(transformed_iris_test_spark_df)}")

iris_train_spark_df2 = iris_train_spark_df.withColumn(
    "validationIndicatorCol", rand(1) > 0.7
)

# train xgboost classifier model with validation dataset
xgb_classifier2 = SparkXGBClassifier(
    max_depth=5, validation_indicator_col="validationIndicatorCol"
)
xgb_classifier_model2 = xgb_classifier2.fit(iris_train_spark_df2)
transformed_iris_test_spark_df2 = xgb_classifier_model2.transform(iris_test_spark_df)
print(
    f"classifier2 f1={classifier_evaluator.evaluate(transformed_iris_test_spark_df2)}"
)

spark.stop()



import json
from typing import Any, Dict, Type

import numpy as np
import pytest

import xgboost as xgb
from xgboost import testing as tm
from xgboost.testing.data import run_base_margin_info
from xgboost.testing.utils import assert_allclose

cp = pytest.importorskip("cupy")


def test_array_interface() -> None:
    arr = cp.array([[1, 2, 3, 4], [1, 2, 3, 4]])
    i_arr = arr.__cuda_array_interface__
    i_arr = json.loads(json.dumps(i_arr))
    ret = xgb.core.from_array_interface(i_arr)
    np.testing.assert_equal(cp.asnumpy(arr), cp.asnumpy(ret))


def dmatrix_from_cupy(
    input_type: Any, DMatrixT: Type[xgb.DMatrix], missing: float = np.nan
) -> xgb.DMatrix:
    """Test constructing DMatrix from cupy"""
    kRows = 80
    kCols = 3

    np_X = np.random.randn(kRows, kCols).astype(dtype=input_type)
    X = cp.array(np_X)
    X[5, 0] = missing
    X[3, 1] = missing
    y = cp.random.randn(kRows).astype(dtype=input_type)
    dtrain = DMatrixT(X, missing=missing, label=y)
    assert dtrain.num_col() == kCols
    assert dtrain.num_row() == kRows

    if DMatrixT is xgb.QuantileDMatrix:
        # Slice is not supported by QuantileDMatrix
        with pytest.raises(xgb.core.XGBoostError):
            dtrain.slice(rindex=[0, 1, 2])
            dtrain.slice(rindex=[0, 1, 2])
    else:
        dtrain.slice(rindex=[0, 1, 2])
        dtrain.slice(rindex=[0, 1, 2])

    return dtrain


def _test_from_cupy(DMatrixT: Type[xgb.DMatrix]) -> None:
    """Test constructing DMatrix from cupy"""
    dmatrix_from_cupy(np.float16, DMatrixT, np.nan)
    dmatrix_from_cupy(np.float32, DMatrixT, np.nan)
    dmatrix_from_cupy(np.float64, DMatrixT, np.nan)

    dmatrix_from_cupy(np.uint8, DMatrixT, 2)
    dmatrix_from_cupy(np.uint32, DMatrixT, 3)
    dmatrix_from_cupy(np.uint64, DMatrixT, 4)

    dmatrix_from_cupy(np.int8, DMatrixT, 2)
    dmatrix_from_cupy(np.int32, DMatrixT, -2)
    dmatrix_from_cupy(np.int64, DMatrixT, -3)

    with pytest.raises(ValueError):
        X = cp.random.randn(2, 2, dtype="float32")
        y = cp.random.randn(2, 2, 3, dtype="float32")
        DMatrixT(X, label=y)


def _test_cupy_training(DMatrixT: Type[xgb.DMatrix]) -> None:
    np.random.seed(1)
    cp.random.seed(np.uint64(1))
    X = cp.random.randn(50, 10, dtype="float32")
    y = cp.random.randn(50, dtype="float32")
    weights = np.random.random(50) + 1
    cupy_weights = cp.array(weights)
    base_margin = np.random.random(50)
    cupy_base_margin = cp.array(base_margin)

    evals_result_cupy: Dict[str, Any] = {}
    dtrain_cp = DMatrixT(X, y, weight=cupy_weights, base_margin=cupy_base_margin)
    params = {"tree_method": "hist", "device": "cuda:0"}
    xgb.train(
        params, dtrain_cp, evals=[(dtrain_cp, "train")], evals_result=evals_result_cupy
    )
    evals_result_np: Dict[str, Any] = {}
    dtrain_np = xgb.DMatrix(
        cp.asnumpy(X), cp.asnumpy(y), weight=weights, base_margin=base_margin
    )
    xgb.train(
        params, dtrain_np, evals=[(dtrain_np, "train")], evals_result=evals_result_np
    )
    assert np.array_equal(
        evals_result_cupy["train"]["rmse"], evals_result_np["train"]["rmse"]
    )


def _test_cupy_metainfo(DMatrixT: Type[xgb.DMatrix]) -> None:
    n = 100
    X = np.random.random((n, 2))
    dmat_cupy = DMatrixT(cp.array(X))
    dmat = xgb.DMatrix(X)
    floats = np.random.random(n)
    uints = np.array([4, 2, 8]).astype("uint32")
    cupy_floats = cp.array(floats)
    cupy_uints = cp.array(uints)
    dmat.set_float_info("weight", floats)
    dmat.set_float_info("label", floats)
    dmat.set_float_info("base_margin", floats)
    dmat.set_uint_info("group", uints)
    dmat_cupy.set_info(weight=cupy_floats)
    dmat_cupy.set_info(label=cupy_floats)
    dmat_cupy.set_info(base_margin=cupy_floats)
    dmat_cupy.set_info(group=cupy_uints)

    # Test setting info with cupy
    assert_allclose("cuda", dmat.get_weight(), dmat_cupy.get_weight())
    assert_allclose("cuda", dmat.get_label(), dmat_cupy.get_label())
    assert_allclose("cuda", dmat.get_base_margin(), dmat_cupy.get_base_margin())
    assert np.array_equal(
        dmat.get_uint_info("group_ptr"), dmat_cupy.get_uint_info("group_ptr")
    )

    run_base_margin_info(cp.asarray, DMatrixT, "cuda")

    dmat_cupy = DMatrixT(cp.array(X))
    y = dmat_cupy.get_label()
    assert y.size == 0
    y = cp.array(X)
    dmat_cupy.set_label(y)
    y1 = dmat_cupy.get_label()
    assert_allclose("cuda", y1, y)


@pytest.mark.skipif(**tm.no_cupy())
@pytest.mark.skipif(**tm.no_sklearn())
def test_cupy_training_with_sklearn() -> None:
    np.random.seed(1)
    cp.random.seed(np.uint64(1))
    X = cp.random.randn(50, 10, dtype="float32")
    y = (cp.random.randn(50, dtype="float32") > 0).astype("int8")
    weights = np.random.random(50) + 1
    cupy_weights = cp.array(weights)
    base_margin = np.random.random(50)
    cupy_base_margin = cp.array(base_margin)

    clf = xgb.XGBClassifier(tree_method="hist", device="cuda:0")
    clf.fit(
        X,
        y,
        sample_weight=cupy_weights,
        base_margin=cupy_base_margin,
        eval_set=[(X, y)],
    )
    pred = clf.predict(X)
    assert np.array_equal(np.unique(pred), np.array([0, 1]))


class TestFromCupy:
    """Tests for constructing DMatrix from data structure conforming Apache
    Arrow specification."""

    @pytest.mark.skipif(**tm.no_cupy())
    def test_simple_dmat_from_cupy(self) -> None:
        _test_from_cupy(xgb.DMatrix)

    @pytest.mark.skipif(**tm.no_cupy())
    def test_quantile_dmat_from_cupy(self) -> None:
        _test_from_cupy(xgb.QuantileDMatrix)

    @pytest.mark.skipif(**tm.no_cupy())
    def test_cupy_training_quantile_dmat(self) -> None:
        _test_cupy_training(xgb.QuantileDMatrix)

    @pytest.mark.skipif(**tm.no_cupy())
    def test_cupy_training_simple_dmat(self) -> None:
        _test_cupy_training(xgb.DMatrix)

    @pytest.mark.skipif(**tm.no_cupy())
    def test_cupy_metainfo_simple_dmat(self) -> None:
        _test_cupy_metainfo(xgb.DMatrix)

    @pytest.mark.skipif(**tm.no_cupy())
    def test_cupy_metainfo_quantile_dmat(self) -> None:
        _test_cupy_metainfo(xgb.QuantileDMatrix)

    @pytest.mark.skipif(**tm.no_cupy())
    def test_dlpack_simple_dmat(self) -> None:
        n = 100
        X = cp.random.random((n, 2))
        capsule = X.__dlpack__()
        xgb.DMatrix(capsule)

    @pytest.mark.skipif(**tm.no_cupy())
    def test_cupy_categorical(self) -> None:
        n_features = 10
        X, y = tm.make_categorical(10, n_features, n_categories=4, onehot=False)
        X = cp.asarray(X.values.astype(cp.float32))
        y = cp.array(y)
        feature_types = ["c"] * n_features

        assert isinstance(X, cp.ndarray)
        Xy = xgb.DMatrix(X, y, feature_types=feature_types)
        np.testing.assert_equal(np.array(Xy.feature_types), np.array(feature_types))

    @pytest.mark.skipif(**tm.no_cupy())
    def test_dlpack_quantile_dmat(self) -> None:
        n = 100
        X = cp.random.random((n, 2))
        m = xgb.QuantileDMatrix(X.__dlpack__())

        with pytest.raises(
            xgb.core.XGBoostError, match="Slicing DMatrix is not supported"
        ):
            m.slice(rindex=[0, 1, 2])

    @pytest.mark.skipif(**tm.no_cupy())
    def test_qid(self) -> None:
        rng = cp.random.RandomState(np.uint64(1994))
        rows = 100
        cols = 10
        X, y = rng.randn(rows, cols), rng.randn(rows)
        qid = rng.randint(low=0, high=10, size=rows, dtype=np.uint32)
        qid = cp.sort(qid)

        Xy = xgb.DMatrix(X, y)
        Xy.set_info(qid=qid)
        group_ptr = Xy.get_uint_info("group_ptr")
        assert group_ptr[0] == 0
        assert group_ptr[-1] == rows

    @pytest.mark.skipif(**tm.no_cupy())
    @pytest.mark.mgpu
    def test_specified_device(self) -> None:
        cp.cuda.runtime.setDevice(0)
        dtrain = dmatrix_from_cupy(np.float32, xgb.QuantileDMatrix, np.nan)
        with pytest.raises(xgb.core.XGBoostError, match="Invalid device ordinal"):
            xgb.train(
                {"tree_method": "hist", "device": "cuda:1"}, dtrain, num_boost_round=10
            )



"""Tests for training continuation."""

import json
from typing import Any, Dict, TypeVar

import numpy as np
import pytest
from hypothesis import strategies

import xgboost as xgb


# pylint: disable=too-many-locals
def run_training_continuation_model_output(device: str, tree_method: str) -> None:
    """Run training continuation test."""
    datasets = pytest.importorskip("sklearn.datasets")
    n_samples = 64
    n_features = 32
    X, y = datasets.make_regression(n_samples, n_features, random_state=1)

    dtrain = xgb.DMatrix(X, y)
    params = {
        "tree_method": tree_method,
        "max_depth": "2",
        "gamma": "0.1",
        "alpha": "0.01",
        "device": device,
    }
    bst_0 = xgb.train(params, dtrain, num_boost_round=64)
    dump_0 = bst_0.get_dump(dump_format="json")

    bst_1 = xgb.train(params, dtrain, num_boost_round=32)
    bst_1 = xgb.train(params, dtrain, num_boost_round=32, xgb_model=bst_1)
    dump_1 = bst_1.get_dump(dump_format="json")

    T = TypeVar("T", Dict[str, Any], float, str, int, list)

    def recursive_compare(obj_0: T, obj_1: T) -> None:
        if isinstance(obj_0, float):
            assert np.isclose(obj_0, obj_1, atol=1e-6)
        elif isinstance(obj_0, str):
            assert obj_0 == obj_1
        elif isinstance(obj_0, int):
            assert obj_0 == obj_1
        elif isinstance(obj_0, dict):
            for i in range(len(obj_0.items())):
                assert list(obj_0.keys())[i] == list(obj_1.keys())[i]
                if list(obj_0.keys())[i] != "missing":
                    recursive_compare(list(obj_0.values()), list(obj_1.values()))
        else:
            for i, lhs in enumerate(obj_0):
                rhs = obj_1[i]
                recursive_compare(lhs, rhs)

    assert len(dump_0) == len(dump_1)

    for i, lhs in enumerate(dump_0):
        obj_0 = json.loads(lhs)
        obj_1 = json.loads(dump_1[i])
        recursive_compare(obj_0, obj_1)


# pylint: disable=too-many-arguments
def run_training_continuation_determinism(
    *,
    device: str,
    tree_method: str,
    subsample: float,
    sampling_method: str,
    rate_drop: float,
    colsample_bylevel: float,
    num_class: int,
    seed_per_iteration: bool,
) -> None:
    """Check that 2-session training (4+4 iters) equals single-session (8 iters)."""
    datasets = pytest.importorskip("sklearn.datasets")

    n_samples = 128
    n_features = 16
    total_rounds = 8
    split_at = 4

    if num_class > 1:
        X, y = datasets.make_classification(
            n_samples=n_samples,
            n_features=n_features,
            n_informative=6,
            n_classes=num_class,
            random_state=42,
        )
        objective = "multi:softprob"
    else:
        X, y = datasets.make_regression(
            n_samples=n_samples, n_features=n_features, random_state=42
        )
        objective = "reg:squarederror"

    dtrain = xgb.DMatrix(X, y)

    params: Dict[str, Any] = {
        "device": device,
        "tree_method": tree_method,
        "max_depth": 4,
        "objective": objective,
        "subsample": subsample,
        "sampling_method": sampling_method,
        "rate_drop": rate_drop,
        "colsample_bylevel": colsample_bylevel,
        "seed_per_iteration": seed_per_iteration,
    }
    if num_class > 1:
        params["num_class"] = num_class

    bst_single = xgb.train(params, dtrain, num_boost_round=total_rounds)

    bst_first = xgb.train(params, dtrain, num_boost_round=split_at)
    bst_continued = xgb.train(
        params, dtrain, num_boost_round=total_rounds - split_at, xgb_model=bst_first
    )

    config_single = json.loads(bst_single.save_config())
    config_cont = json.loads(bst_continued.save_config())

    rng_single = config_single["learner"]["generic_param"]["rng_state"]
    rng_cont = config_cont["learner"]["generic_param"]["rng_state"]
    assert rng_single == rng_cont, "RNG states diverged between single and continued."

    pred_single = bst_single.predict(dtrain)
    pred_cont = bst_continued.predict(dtrain)
    np.testing.assert_allclose(pred_single, pred_cont)


def make_determinism_strategy(tree_methods: list[str]) -> "strategies.SearchStrategy":
    """Hypothesis strategy for testing training continuation with sampling."""
    strategy = strategies.fixed_dictionaries(
        {
            "subsample": strategies.sampled_from([0.5, 1.0]),
            "sampling_method": strategies.sampled_from(["uniform", "gradient_based"]),
            "rate_drop": strategies.sampled_from([0.0, 0.5]),
            "colsample_bylevel": strategies.sampled_from([0.5, 1.0]),
            "tree_method": strategies.sampled_from(tree_methods),
            "num_class": strategies.sampled_from([1, 3]),
            "seed_per_iteration": strategies.booleans(),
        }
    ).filter(
        lambda x: (
            not (
                x["sampling_method"] == "gradient_based" and x["tree_method"] != "hist"
            )
        )
    )
    return strategy



import json
from typing import TYPE_CHECKING, Any, Callable, Dict, Type

import numpy as np
import pytest
import xgboost as xgb
from xgboost import testing as tm
from xgboost.compat import is_dataframe
from xgboost.testing.data import run_base_margin_info
from xgboost.testing.utils import assert_allclose

if TYPE_CHECKING:
    import cudf
else:
    cudf = pytest.importorskip("cudf")


def test_type_check() -> None:
    df = cudf.DataFrame([[1, 2.0], [2, 3.0]], columns=["a", "b"])
    assert is_dataframe(df)
    assert is_dataframe(df.a)


def dmatrix_from_cudf(
    input_type: Any, DMatrixT: Type[xgb.DMatrix], missing: float = np.nan
) -> None:
    """Test constructing DMatrix from cudf"""
    import pandas as pd

    kRows = 80
    kCols = 3

    na = np.random.randn(kRows, kCols)
    na[:, 0:2] = na[:, 0:2].astype(input_type)

    na[5, 0] = missing
    na[3, 1] = missing

    pa = pd.DataFrame({"0": na[:, 0], "1": na[:, 1], "2": na[:, 2].astype(np.int32)})

    np_label = np.random.randn(kRows).astype(input_type)
    pa_label = pd.DataFrame(np_label)

    cd = cudf.from_pandas(pa)
    cd_label = cudf.from_pandas(pa_label).iloc[:, 0]

    dtrain = DMatrixT(cd, missing=missing, label=cd_label)
    assert dtrain.num_col() == kCols
    assert dtrain.num_row() == kRows


def _test_from_cudf(DMatrixT: Type[xgb.DMatrix]) -> None:
    """Test constructing DMatrix from cudf"""
    dmatrix_from_cudf(np.float32, DMatrixT, np.nan)
    dmatrix_from_cudf(np.float64, DMatrixT, np.nan)

    dmatrix_from_cudf(np.int8, DMatrixT, 2)
    dmatrix_from_cudf(np.int32, DMatrixT, -2)
    dmatrix_from_cudf(np.int64, DMatrixT, -3)

    cd = cudf.DataFrame({"x": [1, 2, 3], "y": [0.1, 0.2, 0.3]})
    dtrain = DMatrixT(cd)

    assert dtrain.feature_names == ["x", "y"]
    assert dtrain.feature_types == ["int", "float"]

    series = cudf.DataFrame({"x": [1, 2, 3]}).iloc[:, 0]
    assert isinstance(series, cudf.Series)
    dtrain = DMatrixT(series)

    assert dtrain.feature_names == ["x"]
    assert dtrain.feature_types == ["int"]

    with pytest.raises(ValueError, match=r".*multi.*"):
        dtrain = DMatrixT(cd, label=cd)
        xgb.train(
            {"tree_method": "hist", "device": "cuda", "objective": "multi:softprob"},
            dtrain,
        )

    # Test when number of elements is less than 8
    X = cudf.DataFrame({"x": cudf.Series([0, 1, 2, np.nan, 4], dtype=np.int32)})
    dtrain = DMatrixT(X)
    assert dtrain.num_col() == 1
    assert dtrain.num_row() == 5


def _test_cudf_training(DMatrixT: Type[xgb.DMatrix]) -> None:
    import cudf
    import pandas as pd

    np.random.seed(1)
    X = pd.DataFrame(np.random.randn(50, 10))
    y = pd.DataFrame(np.random.randn(50))
    weights = np.random.random(50) + 1.0
    cudf_weights = cudf.DataFrame(pd.DataFrame(weights))
    base_margin = np.random.random(50)
    cudf_base_margin = cudf.DataFrame(pd.DataFrame(base_margin))

    evals_result_cudf: Dict[str, Any] = {}
    dtrain_cudf = DMatrixT(
        cudf.DataFrame(X),
        cudf.DataFrame(y),
        weight=cudf_weights,
        base_margin=cudf_base_margin,
    )
    params = {"device": "cuda", "tree_method": "hist"}
    xgb.train(
        params,
        dtrain_cudf,
        evals=[(dtrain_cudf, "train")],
        evals_result=evals_result_cudf,
    )
    evals_result_np: Dict[str, Any] = {}
    dtrain_np = xgb.DMatrix(X, y, weight=weights, base_margin=base_margin)
    xgb.train(
        params, dtrain_np, evals=[(dtrain_np, "train")], evals_result=evals_result_np
    )
    assert np.array_equal(
        evals_result_cudf["train"]["rmse"], evals_result_np["train"]["rmse"]
    )


def _test_cudf_metainfo(DMatrixT: Type[xgb.DMatrix]) -> None:
    import cudf
    import pandas as pd

    n = 100
    X = np.random.random((n, 2))
    dmat_cudf = DMatrixT(cudf.DataFrame(pd.DataFrame(X)))
    dmat = xgb.DMatrix(X)
    floats = np.random.random(n)
    uints = np.array([4, 2, 8]).astype("uint32")
    cudf_floats = cudf.DataFrame(pd.DataFrame(floats))
    cudf_uints = cudf.DataFrame(pd.DataFrame(uints))
    dmat.set_weight(floats)
    dmat.set_label(floats)
    dmat.set_base_margin(floats)
    dmat.set_uint_info("group", uints)
    dmat_cudf.set_info(weight=cudf_floats)
    dmat_cudf.set_info(label=cudf_floats)
    dmat_cudf.set_info(base_margin=cudf_floats)
    dmat_cudf.set_info(group=cudf_uints)

    # Test setting info with cudf DataFrame
    assert_allclose("cuda", dmat.get_weight(), dmat_cudf.get_weight())
    assert_allclose("cuda", dmat.get_label(), dmat_cudf.get_label())
    assert_allclose("cuda", dmat.get_base_margin(), dmat_cudf.get_base_margin())
    assert np.array_equal(
        dmat.get_uint_info("group_ptr"), dmat_cudf.get_uint_info("group_ptr")
    )

    # Test setting info with cudf Series
    dmat_cudf.set_info(weight=cudf_floats[cudf_floats.columns[0]])
    dmat_cudf.set_info(label=cudf_floats[cudf_floats.columns[0]])
    dmat_cudf.set_info(base_margin=cudf_floats[cudf_floats.columns[0]])
    dmat_cudf.set_info(group=cudf_uints[cudf_uints.columns[0]])
    assert_allclose("cuda", dmat.get_weight(), dmat_cudf.get_weight())
    assert_allclose("cuda", dmat.get_label(), dmat_cudf.get_label())
    assert_allclose("cuda", dmat.get_base_margin(), dmat_cudf.get_base_margin())
    assert np.array_equal(
        dmat.get_uint_info("group_ptr"), dmat_cudf.get_uint_info("group_ptr")
    )

    run_base_margin_info(cudf.DataFrame, DMatrixT, "cuda")


class TestFromColumnar:
    """Tests for constructing DMatrix from data structure conforming Apache
    Arrow specification."""

    @pytest.mark.skipif(**tm.no_cudf())
    def test_simple_dmatrix_from_cudf(self) -> None:
        _test_from_cudf(xgb.DMatrix)

    @pytest.mark.skipif(**tm.no_cudf())
    def test_device_dmatrix_from_cudf(self) -> None:
        _test_from_cudf(xgb.QuantileDMatrix)

    @pytest.mark.skipif(**tm.no_cudf())
    def test_cudf_training_simple_dmatrix(self) -> None:
        _test_cudf_training(xgb.DMatrix)

    @pytest.mark.skipif(**tm.no_cudf())
    def test_cudf_training_device_dmatrix(self) -> None:
        _test_cudf_training(xgb.QuantileDMatrix)

    @pytest.mark.skipif(**tm.no_cudf())
    def test_cudf_metainfo_simple_dmatrix(self) -> None:
        _test_cudf_metainfo(xgb.DMatrix)

    @pytest.mark.skipif(**tm.no_cudf())
    def test_cudf_metainfo_device_dmatrix(self) -> None:
        _test_cudf_metainfo(xgb.QuantileDMatrix)

    @pytest.mark.skipif(**tm.no_cudf())
    def test_cudf_categorical(self) -> None:
        import pandas as pd

        n_features = 30
        _X, _y = tm.make_categorical(100, n_features, 17, onehot=False)
        X = cudf.from_pandas(_X)
        y = cudf.from_pandas(pd.DataFrame(_y))

        Xy = xgb.DMatrix(X, y)
        assert Xy.feature_types is not None
        assert len(Xy.feature_types) == X.shape[1]
        assert all(t == "c" for t in Xy.feature_types)

        Xy = xgb.QuantileDMatrix(X, y)
        assert Xy.feature_types is not None
        assert len(Xy.feature_types) == X.shape[1]
        assert all(t == "c" for t in Xy.feature_types)

        # mixed dtypes
        X["0"] = X["0"].astype(np.int64)
        X["2"] = X["2"].astype(np.int64)
        df, _, _ = xgb.data._transform_cudf_df(X, None, None, enable_categorical=True)
        assert X.shape[1] == n_features
        assert isinstance(df.aitfs[0], dict)
        assert isinstance(df.aitfs[1], tuple)
        assert isinstance(df.aitfs[2], dict)

        interfaces_str = df.array_interface()
        interfaces = json.loads(interfaces_str)
        assert len(interfaces) == X.shape[1]

        # test missing value
        X = cudf.DataFrame({"f0": ["a", "b", np.nan]})
        X["f0"] = X["f0"].astype("category")
        df, _, _ = xgb.data._transform_cudf_df(X, None, None, enable_categorical=True)
        for col in df.aitfs:
            assert isinstance(col, tuple)
            assert "mask" in col[1]

        y = [0, 1, 2]
        Xy = xgb.DMatrix(X, y)
        assert Xy.num_row() == 3
        assert Xy.num_col() == 1

        Xy = xgb.QuantileDMatrix(X, y)
        assert Xy.num_row() == 3
        assert Xy.num_col() == 1

        X = X["f0"]
        Xy = xgb.DMatrix(X, y)
        assert Xy.num_row() == 3
        assert Xy.num_col() == 1


@pytest.mark.skipif(**tm.no_cudf())
@pytest.mark.skipif(**tm.no_cupy())
@pytest.mark.skipif(**tm.no_sklearn())
@pytest.mark.skipif(**tm.no_pandas())
def test_cudf_training_with_sklearn() -> None:
    import cudf
    import pandas as pd

    np.random.seed(1)
    X = pd.DataFrame(np.random.randn(50, 10))
    y = pd.DataFrame((np.random.randn(50) > 0).astype(np.int8))
    weights = np.random.random(50) + 1.0
    cudf_weights = cudf.DataFrame(pd.DataFrame(weights))
    base_margin = np.random.random(50)
    cudf_base_margin = cudf.DataFrame(pd.DataFrame(base_margin))

    X_cudf = cudf.DataFrame(X)
    y_cudf = cudf.DataFrame(y)
    y_cudf_series = cudf.Series(data=y.iloc[:, 0])

    for y_obj in [y_cudf, y_cudf_series]:
        clf = xgb.XGBClassifier(tree_method="hist", device="cuda:0")
        clf.fit(
            X_cudf,
            y_obj,
            sample_weight=cudf_weights,
            base_margin=cudf_base_margin,
            eval_set=[(X_cudf, y_obj)],
        )
        pred = clf.predict(X_cudf)
        assert np.array_equal(np.unique(pred), np.array([0, 1]))


class IterForDMatrixTest(xgb.core.DataIter):
    """A data iterator for XGBoost DMatrix.

    `reset` and `next` are required for any data iterator, other functions here
    are utilites for demonstration's purpose.

    """

    ROWS_PER_BATCH = 100  # data is splited by rows
    BATCHES = 16

    def __init__(self, categorical: bool) -> None:
        """Generate some random data for demostration.

        Actual data can be anything that is currently supported by XGBoost.
        """
        self.rows = self.ROWS_PER_BATCH

        if categorical:
            self._data = []
            self._labels = []
            for i in range(self.BATCHES):
                X, y = tm.make_categorical(self.ROWS_PER_BATCH, 4, 13, onehot=False)
                self._data.append(cudf.from_pandas(X))
                self._labels.append(y)
        else:
            rng = np.random.RandomState(1994)
            self._data = [
                cudf.DataFrame(
                    {
                        "a": rng.randn(self.ROWS_PER_BATCH),
                        "b": rng.randn(self.ROWS_PER_BATCH),
                    }
                )
            ] * self.BATCHES
            self._labels = [rng.randn(self.rows)] * self.BATCHES

        self.it = 0  # set iterator to 0
        super().__init__(cache_prefix=None)

    def as_array(self) -> "cudf.DataFrame":
        return cudf.concat(self._data)

    def as_array_labels(self) -> np.ndarray:
        return np.concatenate(self._labels)

    def data(self) -> "cudf.DataFrame":
        """Utility function for obtaining current batch of data."""
        return self._data[self.it]

    def labels(self) -> Any:
        """Utility function for obtaining current batch of label."""
        return self._labels[self.it]

    def reset(self) -> None:
        """Reset the iterator"""
        self.it = 0

    def next(self, input_data: Callable) -> bool:
        """Yield next batch of data"""
        if self.it == len(self._data):
            # Return False when there's no more batch.
            return False
        input_data(data=self.data(), label=self.labels())
        self.it += 1
        return True


@pytest.mark.skipif(**tm.no_cudf())
@pytest.mark.parametrize("enable_categorical", [True, False])
def test_from_cudf_iter(enable_categorical: bool) -> None:
    rounds = 100
    it = IterForDMatrixTest(enable_categorical)
    params = {"tree_method": "hist", "device": "cuda"}

    # Use iterator
    m_it = xgb.QuantileDMatrix(it, enable_categorical=enable_categorical)
    reg_with_it = xgb.train(params, m_it, num_boost_round=rounds)

    X = it.as_array()
    y = it.as_array_labels()

    m = xgb.DMatrix(X, y, enable_categorical=enable_categorical)

    assert m_it.num_col() == m.num_col()
    assert m_it.num_row() == m.num_row()

    reg = xgb.train(params, m, num_boost_round=rounds)

    predict = reg.predict(m)
    predict_with_it = reg_with_it.predict(m_it)
    np.testing.assert_allclose(predict_with_it, predict)


def test_invalid_meta() -> None:
    df = cudf.DataFrame({"f0": [0, 1, 2], "f1": [2, 3, 4], "y": [None, 1, 2]})
    y = df["y"]
    X = df.drop(["y"], axis=1)
    with pytest.raises(ValueError, match="Missing value"):
        xgb.DMatrix(X, y)
    with pytest.raises(ValueError, match="Missing value"):
        xgb.QuantileDMatrix(X, y)
    y = X.copy()
    y.iloc[0, 0] = None
    # check by the cuDF->cupy converter.
    with pytest.raises(ValueError, match="Label contains NaN"):
        xgb.DMatrix(X, y)
    with pytest.raises(ValueError, match="Label contains NaN"):
        xgb.QuantileDMatrix(X, y)



from typing import Type

import numpy as np
import pytest
import xgboost as xgb
from xgboost import testing as tm
from xgboost.compat import is_dataframe
from xgboost.core import DataSplitMode
from xgboost.testing.data import pd_arrow_dtypes, pd_dtypes, run_base_margin_info
from xgboost.testing.utils import predictor_equal

try:
    import pandas as pd
except ImportError:
    pass


pytestmark = pytest.mark.skipif(**tm.no_pandas())


dpath = "demo/data/"
rng = np.random.RandomState(1994)


def test_type_check() -> None:
    df = pd.DataFrame([[1, 2.0], [2, 3.0]], columns=["a", "b"])
    assert is_dataframe(df)
    assert is_dataframe(df.a)


class TestPandas:
    def test_pandas(self, data_split_mode=DataSplitMode.ROW):
        world_size = xgb.collective.get_world_size()
        df = pd.DataFrame([[1, 2.0, True], [2, 3.0, False]], columns=["a", "b", "c"])
        assert is_dataframe(df)
        assert is_dataframe(df.a)
        dm = xgb.DMatrix(df, label=pd.Series([1, 2]), data_split_mode=data_split_mode)
        assert dm.num_row() == 2
        if data_split_mode == DataSplitMode.ROW:
            assert dm.feature_names == ["a", "b", "c"]
            assert dm.feature_types == ["int", "float", "i"]
            assert dm.num_col() == 3
        else:
            assert dm.feature_names == tm.column_split_feature_names(
                ["a", "b", "c"], world_size
            )
            assert dm.feature_types == ["int", "float", "i"] * world_size
            assert dm.num_col() == 3 * world_size
        np.testing.assert_array_equal(dm.get_label(), np.array([1, 2]))

        # overwrite feature_names and feature_types
        dm = xgb.DMatrix(
            df,
            label=pd.Series([1, 2]),
            feature_names=["x", "y", "z"],
            feature_types=["q", "q", "q"],
            data_split_mode=data_split_mode,
        )
        assert dm.num_row() == 2
        if data_split_mode == DataSplitMode.ROW:
            assert dm.feature_names == ["x", "y", "z"]
            assert dm.feature_types == ["q", "q", "q"]
            assert dm.num_col() == 3
        else:
            assert dm.feature_names == tm.column_split_feature_names(
                ["x", "y", "z"], world_size
            )
            assert dm.feature_types == ["q", "q", "q"] * world_size
            assert dm.num_col() == 3 * world_size

        # incorrect dtypes
        df = pd.DataFrame([[1, 2.0, "x"], [2, 3.0, "y"]], columns=["a", "b", "c"])
        with pytest.raises(ValueError):
            xgb.DMatrix(df, data_split_mode=data_split_mode)

        # numeric columns
        df = pd.DataFrame([[1, 2.0, True], [2, 3.0, False]])
        dm = xgb.DMatrix(df, label=pd.Series([1, 2]), data_split_mode=data_split_mode)
        assert dm.num_row() == 2
        if data_split_mode == DataSplitMode.ROW:
            assert dm.feature_names == ["0", "1", "2"]
            assert dm.feature_types == ["int", "float", "i"]
            assert dm.num_col() == 3
        else:
            assert dm.feature_names == tm.column_split_feature_names(
                ["0", "1", "2"], world_size
            )
            assert dm.feature_types == ["int", "float", "i"] * world_size
            assert dm.num_col() == 3 * world_size
        np.testing.assert_array_equal(dm.get_label(), np.array([1, 2]))

        df = pd.DataFrame([[1, 2.0, 1], [2, 3.0, 1]], columns=[4, 5, 6])
        dm = xgb.DMatrix(df, label=pd.Series([1, 2]), data_split_mode=data_split_mode)
        assert dm.num_row() == 2
        if data_split_mode == DataSplitMode.ROW:
            assert dm.feature_names == ["4", "5", "6"]
            assert dm.feature_types == ["int", "float", "int"]
            assert dm.num_col() == 3
        else:
            assert dm.feature_names == tm.column_split_feature_names(
                ["4", "5", "6"], world_size
            )
            assert dm.feature_types == ["int", "float", "int"] * world_size
            assert dm.num_col() == 3 * world_size

        df = pd.DataFrame({"A": ["X", "Y", "Z"], "B": [1, 2, 3]})
        dummies = pd.get_dummies(df)
        #    B  A_X  A_Y  A_Z
        # 0  1    1    0    0
        # 1  2    0    1    0
        # 2  3    0    0    1
        result, _, _ = xgb.data._transform_pandas_df(dummies, enable_categorical=False)
        exp = np.array(
            [[1.0, 1.0, 0.0, 0.0], [2.0, 0.0, 1.0, 0.0], [3.0, 0.0, 0.0, 1.0]]
        ).T
        np.testing.assert_array_equal(result.columns, exp)
        dm = xgb.DMatrix(dummies, data_split_mode=data_split_mode)
        assert dm.num_row() == 3
        if data_split_mode == DataSplitMode.ROW:
            assert dm.feature_names == ["B", "A_X", "A_Y", "A_Z"]
            if int(pd.__version__[0]) >= 2:
                assert dm.feature_types == ["int", "i", "i", "i"]
            else:
                assert dm.feature_types == ["int", "int", "int", "int"]
            assert dm.num_col() == 4
        else:
            assert dm.feature_names == tm.column_split_feature_names(
                ["B", "A_X", "A_Y", "A_Z"], world_size
            )
            if int(pd.__version__[0]) >= 2:
                assert dm.feature_types == ["int", "i", "i", "i"] * world_size
            else:
                assert dm.feature_types == ["int", "int", "int", "int"] * world_size
            assert dm.num_col() == 4 * world_size

        df = pd.DataFrame({"A=1": [1, 2, 3], "A=2": [4, 5, 6]})
        dm = xgb.DMatrix(df, data_split_mode=data_split_mode)
        assert dm.num_row() == 3
        if data_split_mode == DataSplitMode.ROW:
            assert dm.feature_names == ["A=1", "A=2"]
            assert dm.feature_types == ["int", "int"]
            assert dm.num_col() == 2
        else:
            assert dm.feature_names == tm.column_split_feature_names(
                ["A=1", "A=2"], world_size
            )
            assert dm.feature_types == ["int", "int"] * world_size
            assert dm.num_col() == 2 * world_size

        df_int = pd.DataFrame([[1, 1.1], [2, 2.2]], columns=[9, 10])
        dm_int = xgb.DMatrix(df_int, data_split_mode=data_split_mode)
        df_range = pd.DataFrame([[1, 1.1], [2, 2.2]], columns=range(9, 11, 1))
        dm_range = xgb.DMatrix(df_range, data_split_mode=data_split_mode)
        if data_split_mode == DataSplitMode.ROW:
            assert dm_int.feature_names == ["9", "10"]  # assert not "9 "
        else:
            assert dm_int.feature_names == tm.column_split_feature_names(
                ["9", "10"], world_size
            )
        assert dm_int.feature_names == dm_range.feature_names

        # test MultiIndex as columns
        df = pd.DataFrame(
            [(1, 2, 3, 4, 5, 6), (6, 5, 4, 3, 2, 1)],
            columns=pd.MultiIndex.from_tuples(
                (
                    ("a", 1),
                    ("a", 2),
                    ("a", 3),
                    ("b", 1),
                    ("b", 2),
                    ("b", 3),
                )
            ),
        )
        dm = xgb.DMatrix(df, data_split_mode=data_split_mode)
        assert dm.num_row() == 2
        if data_split_mode == DataSplitMode.ROW:
            assert dm.feature_names == ["a 1", "a 2", "a 3", "b 1", "b 2", "b 3"]
            assert dm.feature_types == ["int", "int", "int", "int", "int", "int"]
            assert dm.num_col() == 6
        else:
            assert dm.feature_names == tm.column_split_feature_names(
                ["a 1", "a 2", "a 3", "b 1", "b 2", "b 3"], world_size
            )
            assert (
                dm.feature_types
                == ["int", "int", "int", "int", "int", "int"] * world_size
            )
            assert dm.num_col() == 6 * world_size

        # test Index as columns
        df = pd.DataFrame([[1, 1.1], [2, 2.2]], columns=pd.Index([1, 2]))
        Xy = xgb.DMatrix(df, data_split_mode=data_split_mode)
        if data_split_mode == DataSplitMode.ROW:
            np.testing.assert_equal(np.array(Xy.feature_names), np.array(["1", "2"]))
        else:
            np.testing.assert_equal(
                np.array(Xy.feature_names),
                np.array(tm.column_split_feature_names(["1", "2"], world_size)),
            )

        # test pandas series
        data_series = pd.Series([1, 2, 3, 4, 5])
        dm = xgb.DMatrix(data_series, data_split_mode=data_split_mode)
        assert dm.num_row() == 5
        if data_split_mode == DataSplitMode.ROW:
            assert dm.num_col() == 1
        else:
            assert dm.num_col() == 1 * world_size

    @pytest.mark.skipif(**tm.no_sklearn())
    def test_multi_target(self) -> None:
        from sklearn.datasets import make_regression

        X, y = make_regression(n_samples=1024, n_features=4, n_targets=3)
        ydf = pd.DataFrame({i: y[:, i] for i in range(y.shape[1])})

        Xy = xgb.DMatrix(X, ydf)
        assert Xy.num_row() == y.shape[0]
        assert Xy.get_label().size == y.shape[0] * y.shape[1]
        Xy = xgb.QuantileDMatrix(X, ydf)
        assert Xy.num_row() == y.shape[0]
        assert Xy.get_label().size == y.shape[0] * y.shape[1]

    def test_slice(self):
        rng = np.random.RandomState(1994)
        rows = 100
        X = rng.randint(3, 7, size=rows)
        X = pd.DataFrame({"f0": X})
        y = rng.randn(rows)
        ridxs = [1, 2, 3, 4, 5, 6]
        m = xgb.DMatrix(X, y)
        sliced = m.slice(ridxs)

        assert m.feature_types == sliced.feature_types

    def test_pandas_categorical(self, data_split_mode=DataSplitMode.ROW):
        world_size = xgb.collective.get_world_size()
        rng = np.random.RandomState(1994)
        rows = 100
        X = rng.randint(3, 7, size=rows)
        X = pd.Series(X, dtype="category")
        X = pd.DataFrame({"f0": X})
        y = rng.randn(rows)
        m = xgb.DMatrix(X, y, data_split_mode=data_split_mode)
        assert m.feature_types[0] == "c"

        X_0 = ["f", "o", "o"]
        X_1 = [4, 3, 2]
        X = pd.DataFrame({"feat_0": X_0, "feat_1": X_1})
        X["feat_0"] = X["feat_0"].astype("category")
        transformed, _, feature_types = xgb.data._transform_pandas_df(
            X, enable_categorical=True
        )

        assert len(transformed.aitfs[0]) == 2

        # test missing value
        X = pd.DataFrame({"f0": ["a", "b", np.nan]})
        X["f0"] = X["f0"].astype("category")
        arr, _, _ = xgb.data._transform_pandas_df(X, enable_categorical=True)
        for c in arr.columns:
            assert not np.any(c == -1.0)

        X = X["f0"]
        y = y[: X.shape[0]]

        Xy = xgb.DMatrix(X, y, data_split_mode=data_split_mode)
        assert Xy.num_row() == 3
        if data_split_mode == DataSplitMode.ROW:
            assert Xy.num_col() == 1
        else:
            assert Xy.num_col() == 1 * world_size

    def test_pandas_sparse(self):
        import pandas as pd

        rows = 100
        X = pd.DataFrame(
            {
                "A": pd.arrays.SparseArray(np.random.randint(0, 10, size=rows)),
                "B": pd.arrays.SparseArray(np.random.randn(rows)),
                "C": pd.arrays.SparseArray(
                    np.random.permutation([True, False] * (rows // 2))
                ),
            }
        )
        y = pd.Series(pd.arrays.SparseArray(np.random.randn(rows)))
        with pytest.warns(UserWarning, match="Sparse arrays from pandas"):
            dtrain = xgb.DMatrix(X, y)
        booster = xgb.train({}, dtrain, num_boost_round=4)
        with pytest.warns(UserWarning, match="Sparse arrays from pandas"):
            predt_sparse = booster.predict(xgb.DMatrix(X))
            predt_dense = booster.predict(xgb.DMatrix(X.sparse.to_dense()))
        np.testing.assert_allclose(predt_sparse, predt_dense)

    def test_pandas_label(
        self, data_split_mode: DataSplitMode = DataSplitMode.ROW
    ) -> None:
        world_size = xgb.collective.get_world_size()
        # label must be a single column
        df = pd.DataFrame({"A": ["X", "Y", "Z"], "B": [1, 2, 3]})
        with pytest.raises(ValueError):
            xgb.data._transform_pandas_df(df, False, None, None, "label")

        # label must be supported dtype
        df = pd.DataFrame({"A": np.array(["a", "b", "c"], dtype=object)})
        with pytest.raises(ValueError):
            xgb.data._transform_pandas_df(df, False, None, None, "label")

        df = pd.DataFrame({"A": np.array([1, 2, 3], dtype=int)})
        result, _, _ = xgb.data._transform_pandas_df(df, False, None, None, "label")
        np.testing.assert_array_equal(
            np.stack(result.columns, axis=1),
            np.array([[1.0], [2.0], [3.0]], dtype=float),
        )
        dm = xgb.DMatrix(
            np.random.randn(3, 2), label=df, data_split_mode=data_split_mode
        )
        assert dm.num_row() == 3
        if data_split_mode == DataSplitMode.ROW:
            assert dm.num_col() == 2
        else:
            assert dm.num_col() == 2 * world_size

    def test_pandas_weight(self, data_split_mode=DataSplitMode.ROW):
        world_size = xgb.collective.get_world_size()
        kRows = 32
        kCols = 8

        X = np.random.randn(kRows, kCols)
        y = np.random.randn(kRows)
        w = np.random.uniform(size=kRows).astype(np.float32)
        w_pd = pd.DataFrame(w)
        data = xgb.DMatrix(X, y, weight=w_pd, data_split_mode=data_split_mode)

        assert data.num_row() == kRows
        if data_split_mode == DataSplitMode.ROW:
            assert data.num_col() == kCols
        else:
            assert data.num_col() == kCols * world_size
        np.testing.assert_array_equal(data.get_weight(), w)

    def test_base_margin(self):
        run_base_margin_info(pd.DataFrame, xgb.DMatrix, "cpu")

    def test_cv_as_pandas(self):
        dm, _ = tm.load_agaricus(__file__)
        params = {
            "max_depth": 2,
            "eta": 1,
            "objective": "binary:logistic",
            "eval_metric": "error",
        }

        cv = xgb.cv(params, dm, num_boost_round=10, nfold=10)
        assert isinstance(cv, pd.DataFrame)
        exp = pd.Index(
            ["test-error-mean", "test-error-std", "train-error-mean", "train-error-std"]
        )
        assert len(cv.columns.intersection(exp)) == 4

        # show progress log (result is the same as above)
        cv = xgb.cv(params, dm, num_boost_round=10, nfold=10, verbose_eval=True)
        assert isinstance(cv, pd.DataFrame)
        exp = pd.Index(
            ["test-error-mean", "test-error-std", "train-error-mean", "train-error-std"]
        )
        assert len(cv.columns.intersection(exp)) == 4
        cv = xgb.cv(
            params, dm, num_boost_round=10, nfold=10, verbose_eval=True, show_stdv=False
        )
        assert isinstance(cv, pd.DataFrame)
        exp = pd.Index(
            ["test-error-mean", "test-error-std", "train-error-mean", "train-error-std"]
        )
        assert len(cv.columns.intersection(exp)) == 4

        params = {
            "max_depth": 2,
            "eta": 1,
            "objective": "binary:logistic",
            "eval_metric": "auc",
        }
        cv = xgb.cv(params, dm, num_boost_round=10, nfold=10, as_pandas=True)
        assert "eval_metric" in params
        assert "auc" in cv.columns[0]

        params = {
            "max_depth": 2,
            "eta": 1,
            "objective": "binary:logistic",
            "eval_metric": ["auc"],
        }
        cv = xgb.cv(params, dm, num_boost_round=10, nfold=10, as_pandas=True)
        assert "eval_metric" in params
        assert "auc" in cv.columns[0]

        params = {
            "max_depth": 2,
            "eta": 1,
            "objective": "binary:logistic",
            "eval_metric": ["auc"],
        }
        cv = xgb.cv(
            params,
            dm,
            num_boost_round=10,
            nfold=10,
            as_pandas=True,
            early_stopping_rounds=1,
        )
        assert "eval_metric" in params
        assert "auc" in cv.columns[0]
        assert cv.shape[0] < 10

        params = {
            "max_depth": 2,
            "eta": 1,
            "objective": "binary:logistic",
        }
        cv = xgb.cv(
            params, dm, num_boost_round=10, nfold=10, as_pandas=True, metrics="auc"
        )
        assert "auc" in cv.columns[0]

        params = {
            "max_depth": 2,
            "eta": 1,
            "objective": "binary:logistic",
        }
        cv = xgb.cv(
            params, dm, num_boost_round=10, nfold=10, as_pandas=True, metrics=["auc"]
        )
        assert "auc" in cv.columns[0]

        params = {
            "max_depth": 2,
            "eta": 1,
            "objective": "binary:logistic",
            "eval_metric": ["auc"],
        }
        cv = xgb.cv(
            params, dm, num_boost_round=10, nfold=10, as_pandas=True, metrics="error"
        )
        assert "eval_metric" in params
        assert "auc" not in cv.columns[0]
        assert "error" in cv.columns[0]

        cv = xgb.cv(
            params, dm, num_boost_round=10, nfold=10, as_pandas=True, metrics=["error"]
        )
        assert "eval_metric" in params
        assert "auc" not in cv.columns[0]
        assert "error" in cv.columns[0]

        params = list(params.items())
        cv = xgb.cv(
            params, dm, num_boost_round=10, nfold=10, as_pandas=True, metrics=["error"]
        )
        assert isinstance(params, list)
        assert "auc" not in cv.columns[0]
        assert "error" in cv.columns[0]

    @pytest.mark.parametrize("DMatrixT", [xgb.DMatrix, xgb.QuantileDMatrix])
    def test_nullable_type(self, DMatrixT) -> None:
        from xgboost.data import is_pd_cat_dtype

        for orig, df in pd_dtypes():
            if hasattr(df.dtypes, "__iter__"):
                enable_categorical = any(is_pd_cat_dtype(dtype) for dtype in df.dtypes)
            else:
                # series
                enable_categorical = is_pd_cat_dtype(df.dtype)

            f0_orig = orig[orig.columns[0]] if isinstance(orig, pd.DataFrame) else orig
            f0 = df[df.columns[0]] if isinstance(df, pd.DataFrame) else df
            y_orig = f0_orig.astype(pd.Float32Dtype()).fillna(0)
            y = f0.astype(pd.Float32Dtype()).fillna(0)

            m_orig = DMatrixT(orig, enable_categorical=enable_categorical, label=y_orig)
            # extension types
            copy = df.copy()
            m_etype = DMatrixT(df, enable_categorical=enable_categorical, label=y)
            # no mutation
            assert df.equals(copy)
            # different from pd.BooleanDtype(), None is converted to False with bool
            if hasattr(orig.dtypes, "__iter__") and any(
                dtype == "bool" for dtype in orig.dtypes
            ):
                assert not predictor_equal(m_orig, m_etype)
            else:
                assert predictor_equal(m_orig, m_etype)

            np.testing.assert_allclose(m_orig.get_label(), m_etype.get_label())
            np.testing.assert_allclose(m_etype.get_label(), y.values.astype(np.float32))

            if isinstance(df, pd.DataFrame):
                f0 = df["f0"]
                with pytest.raises(ValueError, match="Label contains NaN"):
                    xgb.DMatrix(df, f0, enable_categorical=enable_categorical)

    @pytest.mark.skipif(**tm.no_arrow())
    @pytest.mark.parametrize("DMatrixT", [xgb.DMatrix, xgb.QuantileDMatrix])
    def test_pyarrow_type(self, DMatrixT: Type[xgb.DMatrix]) -> None:
        for orig, df in pd_arrow_dtypes():
            f0_orig: pd.Series = orig["f0"]
            f0 = df["f0"]

            if f0.dtype.name.startswith("bool"):
                y = None
                y_orig = None
            else:
                y_orig = f0_orig.fillna(0, inplace=False)
                y = f0.fillna(0, inplace=False)

            m_orig = DMatrixT(orig, label=y_orig)
            m_etype = DMatrixT(df, label=y)

            assert predictor_equal(m_orig, m_etype)
            if y is not None:
                np.testing.assert_allclose(m_orig.get_label(), m_etype.get_label())
                np.testing.assert_allclose(m_etype.get_label(), y.values)

    @pytest.mark.parametrize("DMatrixT", [xgb.DMatrix, xgb.QuantileDMatrix])
    def test_mixed_type(self, DMatrixT: Type[xgb.DMatrix]) -> None:
        f0 = np.arange(0, 4)
        f1 = pd.Series(f0, dtype="int64[pyarrow]")
        f2l = list(f0)
        f2l[0] = pd.NA
        f2 = pd.Series(f2l, dtype=pd.Int64Dtype())

        df = pd.DataFrame({"f0": f0})
        df["f2"] = f2

        m = DMatrixT(df)
        assert m.num_col() == df.shape[1]

        df["f1"] = f1
        m = DMatrixT(df)
        assert m.num_col() == df.shape[1]
        assert m.num_row() == df.shape[0]
        assert m.num_nonmissing() == df.size - 1
        assert m.feature_names == list(map(str, df.columns))
        assert m.feature_types == ["int"] * df.shape[1]

        y = f0
        m.set_info(label=y)
        booster = xgb.train({}, m)
        p0 = booster.inplace_predict(df)
        p1 = booster.predict(m)
        np.testing.assert_allclose(p0, p1)

    @pytest.mark.skipif(tm.is_windows(), reason="Rabit does not run on windows")
    def test_pandas_column_split(self):
        tm.run_with_rabit(
            world_size=3, test_fn=self.test_pandas, data_split_mode=DataSplitMode.COL
        )

    @pytest.mark.skipif(tm.is_windows(), reason="Rabit does not run on windows")
    def test_pandas_categorical_column_split(self):
        tm.run_with_rabit(
            world_size=3,
            test_fn=self.test_pandas_categorical,
            data_split_mode=DataSplitMode.COL,
        )

    def test_pandas_sparse_column_split(self):
        rows = 100
        X = pd.DataFrame(
            {
                "A": pd.arrays.SparseArray(np.random.randint(0, 10, size=rows)),
                "B": pd.arrays.SparseArray(np.random.randn(rows)),
                "C": pd.arrays.SparseArray(
                    np.random.permutation([True, False] * (rows // 2))
                ),
            }
        )
        y = pd.Series(pd.arrays.SparseArray(np.random.randn(rows)))

        match_sparse = r"Column split does not support pandas sparse array\."
        with pytest.raises(ValueError, match=match_sparse):
            xgb.DMatrix(X, y, data_split_mode=DataSplitMode.COL)
        with pytest.raises(ValueError, match=match_sparse):
            xgb.DMatrix(X, data_split_mode=DataSplitMode.COL)

    @pytest.mark.skipif(tm.is_windows(), reason="Rabit does not run on windows")
    def test_pandas_label_column_split(self):
        tm.run_with_rabit(
            world_size=3,
            test_fn=self.test_pandas_label,
            data_split_mode=DataSplitMode.COL,
        )

    @pytest.mark.skipif(tm.is_windows(), reason="Rabit does not run on windows")
    def test_pandas_weight_column_split(self):
        tm.run_with_rabit(
            world_size=3,
            test_fn=self.test_pandas_weight,
            data_split_mode=DataSplitMode.COL,
        )



"""Tests for updaters."""

import json
from functools import partial, update_wrapper
from string import ascii_lowercase
from typing import Any, Dict, List, Union, overload

import numpy as np
import pytest
from sklearn.datasets import make_regression

import xgboost.testing as tm

from ..callback import TrainingCallback
from ..compat import import_cupy
from ..core import (
    Booster,
    DataIter,
    DMatrix,
    ExtMemQuantileDMatrix,
    QuantileDMatrix,
)
from ..data import is_pd_cat_dtype
from ..sklearn import XGBModel, XGBRegressor
from ..training import train
from .data import IteratorForTest, make_batches, make_categorical
from .data_iter import CatIter
from .utils import Device, assert_allclose, non_increasing


@overload
def get_basescore(model: XGBModel) -> List[float]: ...


@overload
def get_basescore(model: Booster) -> List[float]: ...


@overload
def get_basescore(model: Dict[str, Any]) -> List[float]: ...


def get_basescore(
    model: Union[XGBModel, Booster, Dict],
) -> List[float]:
    """Get base score from an XGBoost sklearn estimator."""
    if isinstance(model, XGBModel):
        model = model.get_booster()

    if isinstance(model, dict):
        jintercept = model["learner"]["learner_model_param"]["base_score"]
    else:
        jintercept = json.loads(model.save_config())["learner"]["learner_model_param"][
            "base_score"
        ]
    return json.loads(jintercept)


# pylint: disable=too-many-locals
def check_quantile_loss(tree_method: str, weighted: bool, device: Device) -> None:
    """Test for quantile loss."""
    from sklearn.metrics import mean_pinball_loss

    from xgboost.sklearn import _metric_decorator

    n_samples = 4096
    n_features = 8
    n_estimators = 8

    rng = np.random.RandomState(1994)
    # pylint: disable=unbalanced-tuple-unpacking
    X, y = make_regression(
        n_samples=n_samples,
        n_features=n_features,
        random_state=rng,
    )
    if weighted:
        weight = rng.random(size=n_samples)
    else:
        weight = None

    Xy = QuantileDMatrix(X, y, weight=weight)

    alpha = np.array([0.1, 0.5])
    # non-zero base score can cause floating point difference with GPU predictor.
    # multi-class has small difference than single target in the prediction kernel
    base_score = np.zeros(shape=alpha.shape, dtype=np.float32)
    evals_result: Dict[str, Dict] = {}
    booster_multi = train(
        {
            "objective": "reg:quantileerror",
            "tree_method": tree_method,
            "device": device,
            "quantile_alpha": alpha,
            "base_score": base_score,
        },
        Xy,
        num_boost_round=n_estimators,
        evals=[(Xy, "Train")],
        evals_result=evals_result,
    )
    predt_multi = booster_multi.predict(Xy, strict_shape=True)

    assert non_increasing(evals_result["Train"]["quantile"])
    assert evals_result["Train"]["quantile"][-1] < 20.0
    # check that there's a way to use custom metric and compare the results.
    metrics = [
        _metric_decorator(
            update_wrapper(
                partial(mean_pinball_loss, sample_weight=weight, alpha=alpha[i]),
                mean_pinball_loss,
            )
        )
        for i in range(alpha.size)
    ]

    predts = np.empty(predt_multi.shape)
    for i in range(alpha.shape[0]):
        a = alpha[i]

        booster_i = train(
            {
                "objective": "reg:quantileerror",
                "tree_method": tree_method,
                "device": device,
                "quantile_alpha": a,
                "base_score": base_score[i],
            },
            Xy,
            num_boost_round=n_estimators,
            evals=[(Xy, "Train")],
            custom_metric=metrics[i],
            evals_result=evals_result,
        )
        assert non_increasing(evals_result["Train"]["quantile"])
        assert evals_result["Train"]["quantile"][-1] < 30.0
        np.testing.assert_allclose(
            np.array(evals_result["Train"]["quantile"]),
            np.array(evals_result["Train"]["mean_pinball_loss"]),
            atol=1e-6,
            rtol=1e-6,
        )
        predts[:, i] = booster_i.predict(Xy)

    for i in range(alpha.shape[0]):
        np.testing.assert_allclose(predts[:, i], predt_multi[:, i])


def check_quantile_loss_rf(
    device: Device, tree_method: str, multi_strategy: str
) -> None:
    """Test quantile loss with boosting random forest."""
    # pylint: disable=unbalanced-tuple-unpacking
    X, y = make_regression(n_samples=2048, n_features=16, random_state=2026)
    Xy = DMatrix(X, y)

    def run(params: Dict[str, Any], metric: str) -> None:
        evals_result_0: Dict[str, Dict] = {}
        params["num_parallel_tree"] = 2
        train(
            params,
            Xy,
            num_boost_round=8,
            evals=[(Xy, "Train")],
            evals_result=evals_result_0,
        )

        evals_result_1: Dict[str, Dict] = {}
        params["num_parallel_tree"] = 1
        train(
            params,
            Xy,
            num_boost_round=8,
            evals=[(Xy, "Train")],
            evals_result=evals_result_1,
        )
        # Without subsample, the result should be the same (barring floating point
        # errors).
        np.testing.assert_allclose(
            evals_result_0["Train"][metric], evals_result_1["Train"][metric]
        )
        assert non_increasing(evals_result_0["Train"][metric])

    alpha = np.array([0.1, 0.5, 0.9])
    params = {
        "objective": "reg:quantileerror",
        "tree_method": tree_method,
        "device": device,
        "quantile_alpha": alpha,
        "multi_strategy": multi_strategy,
    }
    run(params, "quantile")

    # Now test with MAE
    params.pop("quantile_alpha")
    params["objective"] = "reg:absoluteerror"
    run(params, "mae")


def check_quantile_loss_extmem(
    n_samples_per_batch: int,
    n_features: int,
    n_batches: int,
    tree_method: str,
    device: str,
) -> None:
    """Check external memory with the quantile objective."""
    it = IteratorForTest(
        *make_batches(n_samples_per_batch, n_features, n_batches, device != "cpu"),
        cache="cache",
        on_host=False,
    )
    Xy_it = DMatrix(it)
    params = {
        "tree_method": tree_method,
        "objective": "reg:quantileerror",
        "device": device,
        "quantile_alpha": [0.2, 0.8],
    }
    booster_it = train(params, Xy_it)
    X, y, w = it.as_arrays()
    Xy = DMatrix(X, y, weight=w)
    booster = train(params, Xy)

    predt_it = booster_it.predict(Xy_it)
    predt = booster.predict(Xy)

    np.testing.assert_allclose(predt, predt_it)


def check_extmem_qdm(  # pylint: disable=too-many-arguments
    n_samples_per_batch: int,
    n_features: int,
    *,
    n_batches: int,
    n_bins: int,
    device: str,
    on_host: bool,
    is_cat: bool,
) -> None:
    """Basic test for the `ExtMemQuantileDMatrix`."""

    if is_cat:
        it: DataIter = CatIter(
            n_samples_per_batch=n_samples_per_batch,
            n_features=n_features,
            n_batches=n_batches,
            n_cats=5,
            sparsity=0.0,
            cat_ratio=0.5,
            onehot=False,
            device=device,
            cache="cache",
        )
    else:
        it = IteratorForTest(
            *make_batches(
                n_samples_per_batch, n_features, n_batches, use_cupy=device != "cpu"
            ),
            cache="cache",
            on_host=on_host,
        )

    Xy_it = ExtMemQuantileDMatrix(it, max_bin=n_bins, enable_categorical=is_cat)
    with pytest.raises(ValueError, match="Only the `hist`"):
        booster_it = train(
            {"device": device, "tree_method": "approx", "max_bin": n_bins},
            Xy_it,
            num_boost_round=8,
        )

    booster_it = train({"device": device, "max_bin": n_bins}, Xy_it, num_boost_round=8)
    if is_cat:
        it = CatIter(
            n_samples_per_batch=n_samples_per_batch,
            n_features=n_features,
            n_batches=n_batches,
            n_cats=5,
            sparsity=0.0,
            cat_ratio=0.5,
            onehot=False,
            device=device,
            cache=None,
        )
    else:
        it = IteratorForTest(
            *make_batches(
                n_samples_per_batch, n_features, n_batches, use_cupy=device != "cpu"
            ),
            cache=None,
        )
    Xy = QuantileDMatrix(it, max_bin=n_bins, enable_categorical=is_cat)
    booster = train({"device": device, "max_bin": n_bins}, Xy, num_boost_round=8)

    cut_it = Xy_it.get_quantile_cut()
    cut = Xy.get_quantile_cut()
    np.testing.assert_allclose(cut_it[0], cut[0])
    np.testing.assert_allclose(cut_it[1], cut[1])

    predt_it = booster_it.predict(Xy_it)
    predt = booster.predict(Xy)
    np.testing.assert_allclose(predt_it, predt)


def check_cut(
    n_entries: int, indptr: np.ndarray, data: np.ndarray, dtypes: Any
) -> None:
    """Check the cut values."""
    assert data.shape[0] == indptr[-1]
    assert data.shape[0] == n_entries

    assert indptr.dtype == np.uint64
    for i in range(1, indptr.size):
        beg = int(indptr[i - 1])
        end = int(indptr[i])
        for j in range(beg + 1, end):
            assert data[j] > data[j - 1]
            if is_pd_cat_dtype(dtypes.iloc[i - 1]):
                assert data[j] == data[j - 1] + 1


def check_get_quantile_cut_device(tree_method: str, use_cupy: bool) -> None:
    """Check with optional cupy."""
    import pandas as pd

    n_samples = 1024
    n_features = 14
    max_bin = 16
    dtypes = pd.Series([np.float32] * n_features)

    # numerical
    X, y, w = tm.make_regression(n_samples, n_features, use_cupy=use_cupy)
    # - qdm
    Xyw: DMatrix = QuantileDMatrix(X, y, weight=w, max_bin=max_bin)
    indptr, data = Xyw.get_quantile_cut()
    check_cut((max_bin + 1) * n_features, indptr, data, dtypes)
    # - dm
    Xyw = DMatrix(X, y, weight=w)
    train({"tree_method": tree_method, "max_bin": max_bin}, Xyw)
    indptr, data = Xyw.get_quantile_cut()
    check_cut((max_bin + 1) * n_features, indptr, data, dtypes)
    # - ext mem
    n_batches = 3
    n_samples_per_batch = 256
    it = IteratorForTest(
        *make_batches(n_samples_per_batch, n_features, n_batches, use_cupy),
        cache="cache",
        on_host=False,
    )
    Xy: DMatrix = DMatrix(it)
    train({"tree_method": tree_method, "max_bin": max_bin}, Xyw)
    indptr, data = Xyw.get_quantile_cut()
    check_cut((max_bin + 1) * n_features, indptr, data, dtypes)

    # categorical
    n_categories = 32
    X, y = make_categorical(
        n_samples, n_features, n_categories, onehot=False, sparsity=0.8
    )
    if use_cupy:
        import cudf

        cp = import_cupy()

        X = cudf.from_pandas(X)
        y = cp.array(y)
    # - qdm
    Xy = QuantileDMatrix(X, y, max_bin=max_bin, enable_categorical=True)
    indptr, data = Xy.get_quantile_cut()
    check_cut(n_categories * n_features, indptr, data, X.dtypes)
    # - dm
    Xy = DMatrix(X, y, enable_categorical=True)
    train({"tree_method": tree_method, "max_bin": max_bin}, Xy)
    indptr, data = Xy.get_quantile_cut()
    check_cut(n_categories * n_features, indptr, data, X.dtypes)

    # mixed
    X, y = make_categorical(
        n_samples, n_features, n_categories, onehot=False, sparsity=0.8, cat_ratio=0.5
    )
    n_cat_features = len([0 for dtype in X.dtypes if is_pd_cat_dtype(dtype)])
    n_num_features = n_features - n_cat_features
    n_entries = n_categories * n_cat_features + (max_bin + 1) * n_num_features
    # - qdm
    Xy = QuantileDMatrix(X, y, max_bin=max_bin, enable_categorical=True)
    indptr, data = Xy.get_quantile_cut()
    check_cut(n_entries, indptr, data, X.dtypes)
    # - dm
    Xy = DMatrix(X, y, enable_categorical=True)
    train({"tree_method": tree_method, "max_bin": max_bin}, Xy)
    indptr, data = Xy.get_quantile_cut()
    check_cut(n_entries, indptr, data, X.dtypes)


def check_get_quantile_cut(tree_method: str, device: str) -> None:
    """Check the quantile cut getter."""

    use_cupy = device.startswith("cuda")
    check_get_quantile_cut_device(tree_method, False)
    if use_cupy:
        check_get_quantile_cut_device(tree_method, True)


USE_ONEHOT = np.iinfo(np.int32).max
USE_PART = 1


def _create_dmatrix(  # pylint: disable=too-many-arguments
    n_samples: int,
    n_features: int,
    *,
    n_cats: int,
    device: str,
    sparsity: float,
    tree_method: str,
    onehot: bool,
    extmem: bool,
    enable_categorical: bool,
) -> DMatrix:
    n_batches = max(min(2, n_samples), 1)
    it = CatIter(
        n_samples // n_batches,
        n_features,
        n_batches=n_batches,
        sparsity=sparsity,
        cat_ratio=1.0,
        n_cats=n_cats,
        onehot=onehot,
        device=device,
        cache="cache" if extmem else None,
    )
    if extmem:
        if tree_method == "hist":
            Xy: DMatrix = ExtMemQuantileDMatrix(
                it, enable_categorical=enable_categorical
            )
        elif tree_method == "approx":
            Xy = DMatrix(it, enable_categorical=enable_categorical)
        else:
            raise ValueError(f"tree_method {tree_method} not supported.")
    else:
        cat, label = it.xy()
        Xy = DMatrix(cat, label, enable_categorical=enable_categorical)
    return Xy


def check_categorical_ohe(  # pylint: disable=too-many-arguments
    *,
    rows: int,
    cols: int,
    rounds: int,
    cats: int,
    device: str,
    tree_method: str,
    extmem: bool = False,
    multi_target: bool = False,
) -> None:
    "Test for one-hot encoding with categorical data."
    pd = pytest.importorskip("pandas")

    by_etl_results: Dict[str, Dict[str, List[float]]] = {}
    by_builtin_results: Dict[str, Dict[str, List[float]]] = {}

    parameters: Dict[str, Any] = {
        "tree_method": tree_method,
        # Use one-hot exclusively
        "max_cat_to_onehot": USE_ONEHOT,
        "device": device,
    }

    if multi_target:
        n_targets = 3
        parameters["multi_strategy"] = "multi_output_tree"

        cat, label = make_categorical(
            rows,
            cols,
            n_categories=cats,
            onehot=False,
            sparsity=0.0,
            n_targets=n_targets,
        )

        Xy_onehot = DMatrix(pd.get_dummies(cat), label)
        Xy_cat = DMatrix(cat, label, enable_categorical=True)
    else:
        Xy_onehot = _create_dmatrix(
            rows,
            cols,
            n_cats=cats,
            device=device,
            sparsity=0.0,
            onehot=True,
            tree_method=tree_method,
            extmem=extmem,
            enable_categorical=False,
        )
        Xy_cat = _create_dmatrix(
            rows,
            cols,
            n_cats=cats,
            device=device,
            sparsity=0.0,
            tree_method=tree_method,
            onehot=False,
            extmem=extmem,
            enable_categorical=True,
        )

    train(
        parameters,
        Xy_onehot,
        num_boost_round=rounds,
        evals=[(Xy_onehot, "Train")],
        evals_result=by_etl_results,
    )
    train(
        parameters,
        Xy_cat,
        num_boost_round=rounds,
        evals=[(Xy_cat, "Train")],
        evals_result=by_builtin_results,
    )

    # There are guidelines on how to specify tolerance based on considering output
    # as random variables. But in here the tree construction is extremely sensitive
    # to floating point errors. An 1e-5 error in a histogram bin can lead to an
    # entirely different tree. So even though the test is quite lenient, hypothesis
    # can still pick up falsifying examples from time to time.
    np.testing.assert_allclose(
        np.array(by_etl_results["Train"]["rmse"]),
        np.array(by_builtin_results["Train"]["rmse"]),
        rtol=1e-3,
    )
    assert non_increasing(by_builtin_results["Train"]["rmse"])

    if not multi_target:
        by_grouping: Dict[str, Dict[str, List[float]]] = {}
        # switch to partition-based splits
        parameters["max_cat_to_onehot"] = USE_PART
        parameters["reg_lambda"] = 0
        train(
            parameters,
            Xy_cat,
            num_boost_round=rounds,
            evals=[(Xy_cat, "Train")],
            evals_result=by_grouping,
        )
        rmse_oh = by_builtin_results["Train"]["rmse"]
        rmse_group = by_grouping["Train"]["rmse"]
        # always better or equal to onehot when there's no regularization.
        for a, b in zip(rmse_oh, rmse_group):
            assert a >= b

        parameters["reg_lambda"] = 1.0
        by_grouping = {}
        train(
            parameters,
            Xy_cat,
            num_boost_round=32,
            evals=[(Xy_cat, "Train")],
            evals_result=by_grouping,
        )
        assert non_increasing(by_grouping["Train"]["rmse"]), by_grouping


def check_categorical_missing(  # pylint: disable=too-many-arguments
    rows: int,
    cols: int,
    cats: int,
    *,
    device: Device,
    tree_method: str,
    extmem: bool,
) -> None:
    """Check categorical data with missing values."""
    parameters: Dict[str, Any] = {"tree_method": tree_method, "device": device}
    Xy = _create_dmatrix(
        rows,
        cols,
        n_cats=cats,
        sparsity=0.5,
        device=device,
        tree_method=tree_method,
        onehot=False,
        extmem=extmem,
        enable_categorical=True,
    )
    label = Xy.get_label()

    def run(max_cat_to_onehot: int) -> None:
        # Test with onehot splits
        parameters["max_cat_to_onehot"] = max_cat_to_onehot

        evals_result: Dict[str, Dict] = {}
        booster = train(
            parameters,
            Xy,
            num_boost_round=8,
            evals=[(Xy, "Train")],
            evals_result=evals_result,
        )
        assert non_increasing(evals_result["Train"]["rmse"])
        y_predt = booster.predict(Xy)
        rmse = tm.root_mean_square(label, y_predt)
        assert_allclose(device, rmse, evals_result["Train"]["rmse"][-1], rtol=2e-5)

    # Test with OHE split
    run(USE_ONEHOT)

    # Test with partition-based split
    run(USE_PART)


def run_max_cat(tree_method: str, device: Device) -> None:
    """Test data with size smaller than number of categories."""
    import pandas as pd

    rng = np.random.default_rng(0)
    n_cat = 100
    n = 5

    X = pd.Series(
        ["".join(rng.choice(list(ascii_lowercase), size=3)) for i in range(n_cat)],
        dtype="category",
    )[:n].to_frame()

    reg = XGBRegressor(
        enable_categorical=True,
        tree_method=tree_method,
        device=device,
        n_estimators=10,
    )
    y = pd.Series(range(n))
    reg.fit(X=X, y=y, eval_set=[(X, y)])
    assert non_increasing(reg.evals_result()["validation_0"]["rmse"])


def run_invalid_category(tree_method: str, device: Device) -> None:
    """Test with invalid categorical inputs."""
    rng = np.random.default_rng()
    # too large
    X = rng.integers(low=0, high=4, size=1000).reshape(100, 10)
    y = rng.normal(loc=0, scale=1, size=100)
    X[13, 7] = np.iinfo(np.int32).max + 1

    # Check is performed during sketching.
    Xy = DMatrix(X, y, feature_types=["c"] * 10)
    with pytest.raises(ValueError):
        train({"tree_method": tree_method, "device": device}, Xy)

    X[13, 7] = 16777216
    Xy = DMatrix(X, y, feature_types=["c"] * 10)
    with pytest.raises(ValueError):
        train({"tree_method": tree_method, "device": device}, Xy)

    # mixed positive and negative values
    X = rng.normal(loc=0, scale=1, size=1000).reshape(100, 10)  # type: ignore[assignment]
    y = rng.normal(loc=0, scale=1, size=100)

    Xy = DMatrix(X, y, feature_types=["c"] * 10)
    with pytest.raises(ValueError):
        train({"tree_method": tree_method, "device": device}, Xy)

    if device == "cuda":
        import cupy as cp

        X, y = cp.array(X), cp.array(y)
        with pytest.raises(ValueError):
            QuantileDMatrix(X, y, feature_types=["c"] * 10)


def train_result(
    param: Dict[str, Any], dmat: DMatrix, num_rounds: int
) -> Dict[str, Any]:
    """Get training result from parameters and data."""
    result: Dict[str, Any] = {}
    booster = train(
        param,
        dmat,
        num_rounds,
        evals=[(dmat, "train")],
        verbose_eval=False,
        evals_result=result,
    )
    assert booster.num_features() == dmat.num_col()
    assert booster.num_boosted_rounds() == num_rounds
    assert booster.feature_names == dmat.feature_names
    assert booster.feature_types == dmat.feature_types

    return result


class ResetStrategy(TrainingCallback):
    """Callback for testing multi-output."""

    def after_iteration(self, model: Booster, epoch: int, evals_log: dict) -> bool:
        if epoch % 2 == 0:
            model.set_param({"multi_strategy": "multi_output_tree"})
        else:
            model.set_param({"multi_strategy": "one_output_per_tree"})
        return False



import json
import os
from typing import List, Optional, Tuple, cast

import numpy as np
import pytest

import xgboost as xgb
from xgboost import testing as tm

dpath = tm.data_dir(__file__)


@pytest.fixture(scope="module")
def toy_data() -> Tuple[xgb.DMatrix, np.ndarray, np.ndarray]:
    X = np.array([1, 2, 3, 4, 5]).reshape((-1, 1))
    INF = np.inf
    y_lower = np.array([10, 15, -INF, 30, 100])
    y_upper = np.array([INF, INF, 20, 50, INF])

    dmat = xgb.DMatrix(X)
    dmat.set_float_info("label_lower_bound", y_lower)
    dmat.set_float_info("label_upper_bound", y_upper)
    return dmat, y_lower, y_upper


def test_default_metric(toy_data: Tuple[xgb.DMatrix, np.ndarray, np.ndarray]) -> None:
    Xy, y_lower, y_upper = toy_data

    def run(evals: Optional[list]) -> None:
        # test with or without actual evaluation.
        booster = xgb.train(
            {"objective": "survival:aft", "aft_loss_distribution": "extreme"},
            Xy,
            num_boost_round=1,
            evals=evals,
        )
        config = json.loads(booster.save_config())
        metrics = config["learner"]["metrics"]
        assert len(metrics) == 1
        assert metrics[0]["aft_loss_param"]["aft_loss_distribution"] == "extreme"

        booster = xgb.train(
            {"objective": "survival:aft"},
            Xy,
            num_boost_round=1,
            evals=evals,
        )
        config = json.loads(booster.save_config())
        metrics = config["learner"]["metrics"]
        assert len(metrics) == 1
        assert metrics[0]["aft_loss_param"]["aft_loss_distribution"] == "normal"

    run([(Xy, "Train")])
    run(None)


def test_aft_survival_toy_data(
    toy_data: Tuple[xgb.DMatrix, np.ndarray, np.ndarray]
) -> None:
    # See demo/aft_survival/aft_survival_viz_demo.py
    X = np.array([1, 2, 3, 4, 5]).reshape((-1, 1))
    dmat, y_lower, y_upper = toy_data

    # "Accuracy" = the number of data points whose ranged label (y_lower, y_upper)
    #              includes the corresponding predicted label (y_pred)
    acc_rec = []

    class Callback(xgb.callback.TrainingCallback):
        def __init__(self):
            super().__init__()

        def after_iteration(
            self,
            model: xgb.Booster,
            epoch: int,
            evals_log: xgb.callback.TrainingCallback.EvalsLog,
        ):
            y_pred = model.predict(dmat)
            acc = np.sum(np.logical_and(y_pred >= y_lower, y_pred <= y_upper) / len(X))
            acc_rec.append(acc)
            return False

    evals_result: xgb.callback.TrainingCallback.EvalsLog = {}
    params = {
        "max_depth": 3,
        "objective": "survival:aft",
        "min_child_weight": 0,
        "tree_method": "exact",
    }
    bst = xgb.train(
        params,
        dmat,
        15,
        [(dmat, "train")],
        evals_result=evals_result,
        callbacks=[Callback()],
    )

    nloglik_rec = cast(List[float], evals_result["train"]["aft-nloglik"])
    # AFT metric (negative log likelihood) improve monotonically
    assert all(p >= q for p, q in zip(nloglik_rec, nloglik_rec[:1]))
    # "Accuracy" improve monotonically.
    # Over time, XGBoost model makes predictions that fall within given label ranges.
    assert all(p <= q for p, q in zip(acc_rec, acc_rec[1:]))
    assert acc_rec[-1] == 1.0

    def gather_split_thresholds(tree):
        if "split_condition" in tree:
            return (
                gather_split_thresholds(tree["children"][0])
                | gather_split_thresholds(tree["children"][1])
                | {tree["split_condition"]}
            )
        return set()

    # Only 2.5, 3.5, and 4.5 are used as split thresholds.
    model_json = [json.loads(e) for e in bst.get_dump(dump_format="json")]
    for i, tree in enumerate(model_json):
        assert gather_split_thresholds(tree).issubset({2.5, 3.5, 4.5})


def test_aft_empty_dmatrix():
    X = np.array([]).reshape((0, 2))
    y_lower, y_upper = np.array([]), np.array([])
    dtrain = xgb.DMatrix(X)
    dtrain.set_info(label_lower_bound=y_lower, label_upper_bound=y_upper)
    bst = xgb.train({'objective': 'survival:aft', 'tree_method': 'hist'},
                    dtrain, num_boost_round=2, evals=[(dtrain, 'train')])


@pytest.mark.skipif(**tm.no_pandas())
def test_aft_survival_demo_data():
    import pandas as pd
    df = pd.read_csv(os.path.join(dpath, 'veterans_lung_cancer.csv'))

    y_lower_bound = df['Survival_label_lower_bound']
    y_upper_bound = df['Survival_label_upper_bound']
    X = df.drop(['Survival_label_lower_bound', 'Survival_label_upper_bound'], axis=1)

    dtrain = xgb.DMatrix(X)
    dtrain.set_float_info('label_lower_bound', y_lower_bound)
    dtrain.set_float_info('label_upper_bound', y_upper_bound)

    base_params = {'verbosity': 0,
                   'objective': 'survival:aft',
                   'eval_metric': 'aft-nloglik',
                   'tree_method': 'hist',
                   'learning_rate': 0.05,
                   'aft_loss_distribution_scale': 1.20,
                   'max_depth': 6,
                   'lambda': 0.01,
                   'alpha': 0.02}
    nloglik_rec = {}
    dists = ['normal', 'logistic', 'extreme']
    for dist in dists:
        params = base_params
        params.update({'aft_loss_distribution': dist})
        evals_result = {}
        bst = xgb.train(params, dtrain, num_boost_round=500, evals=[(dtrain, 'train')],
                        evals_result=evals_result)
        nloglik_rec[dist] = evals_result['train']['aft-nloglik']
        # AFT metric (negative log likelihood) improve monotonically
        assert all(p >= q for p, q in zip(nloglik_rec[dist], nloglik_rec[dist][:1]))
    # For this data, normal distribution works the best
    assert nloglik_rec['normal'][-1] < 4.9
    assert nloglik_rec['logistic'][-1] > 4.9
    assert nloglik_rec['extreme'][-1] > 4.9



"""Tests for inference."""

from typing import Type

import numpy as np
from scipy.special import logit  # pylint: disable=no-name-in-module

from ..core import DMatrix
from ..training import train
from .shared import validate_leaf_output
from .updater import get_basescore
from .utils import Device


# pylint: disable=too-many-locals
def run_predict_leaf(device: Device, DMatrixT: Type[DMatrix]) -> np.ndarray:
    """Run tests for leaf index prediction."""
    rows = 100
    cols = 4
    classes = 5
    num_parallel_tree = 4
    num_boost_round = 10
    rng = np.random.RandomState(1994)
    X = rng.randn(rows, cols)
    y = rng.randint(low=0, high=classes, size=rows)

    m = DMatrixT(X, y)
    booster = train(
        {
            "num_parallel_tree": num_parallel_tree,
            "num_class": classes,
            "tree_method": "hist",
        },
        m,
        num_boost_round=num_boost_round,
    )

    booster.set_param({"device": device})
    empty = DMatrixT(np.ones(shape=(0, cols)))
    empty_leaf = booster.predict(empty, pred_leaf=True)
    assert empty_leaf.shape[0] == 0

    leaf = booster.predict(m, pred_leaf=True, strict_shape=True)
    assert leaf.shape[0] == rows
    assert leaf.shape[1] == num_boost_round
    assert leaf.shape[2] == classes
    assert leaf.shape[3] == num_parallel_tree

    validate_leaf_output(leaf, num_parallel_tree)

    n_iters = np.int32(2)
    sliced = booster.predict(
        m,
        pred_leaf=True,
        iteration_range=(0, n_iters),
        strict_shape=True,
    )
    first = sliced[0, ...]

    assert np.prod(first.shape) == classes * num_parallel_tree * n_iters

    # When there's only 1 tree, the output is a 1 dim vector
    booster = train({"tree_method": "hist"}, num_boost_round=1, dtrain=m)
    booster.set_param({"device": device})
    assert booster.predict(m, pred_leaf=True).shape == (rows,)

    return leaf


def run_base_margin_vs_base_score(device: Device) -> None:
    """Test for the relation between score and margin."""
    from sklearn.datasets import make_classification

    intercept = 0.5

    X, y = make_classification(random_state=2025)
    booster = train(
        {"base_score": intercept, "objective": "binary:logistic", "device": device},
        dtrain=DMatrix(X, y),
        num_boost_round=1,
    )
    np.testing.assert_allclose(get_basescore(booster), intercept)
    predt_0 = booster.predict(DMatrix(X, y))

    margin = np.full(y.shape, fill_value=logit(intercept), dtype=np.float32)
    Xy = DMatrix(X, y, base_margin=margin)
    # 0.2 is a dummy value
    booster = train(
        {"base_score": 0.2, "objective": "binary:logistic", "device": device},
        dtrain=Xy,
        num_boost_round=1,
    )
    np.testing.assert_allclose(get_basescore(booster), 0.2)
    predt_1 = booster.predict(Xy)

    np.testing.assert_allclose(predt_0, predt_1)



"""Test model IO with pickle."""

import os
import pickle
import subprocess
from typing import Any, Dict, Tuple, Union

import numpy as np
import pytest

import xgboost as xgb
from xgboost import XGBClassifier
from xgboost import testing as tm

model_path = "./model.pkl"

pytestmark = tm.timeout(30)


def build_dataset() -> Tuple[np.ndarray, np.ndarray]:
    N = 10
    x = np.linspace(0, N * N, N * N)
    x = x.reshape((N, N))
    y = np.linspace(0, N, N)
    return x, y


def save_pickle(bst: Union[xgb.Booster, xgb.XGBModel], path: str) -> None:
    with open(path, "wb") as fd:
        pickle.dump(bst, fd)


def load_pickle(path: str) -> Any:
    with open(path, "rb") as fd:
        bst = pickle.load(fd)
    return bst


class TestPickling:
    args_template = ["pytest", "--verbose", "-s", "--fulltrace"]

    def run_pickling(self, bst: Union[xgb.Booster, xgb.XGBModel]) -> None:
        save_pickle(bst, model_path)
        args = [
            "pytest",
            "--verbose",
            "-s",
            "--fulltrace",
            "./tests/python-gpu/load_pickle.py::TestLoadPickle::test_load_pkl",
        ]
        command = ""
        for arg in args:
            command += arg
            command += " "

        cuda_environment = {"CUDA_VISIBLE_DEVICES": "-1"}
        env = os.environ.copy()
        # Passing new_environment directly to `env' argument results
        # in failure on Windows:
        #    Fatal Python error: _Py_HashRandomization_Init: failed to
        #    get random numbers to initialize Python
        env.update(cuda_environment)

        # Load model in a CPU only environment.
        status = subprocess.call(command, env=env, shell=True)
        assert status == 0
        os.remove(model_path)

    # TODO: This test is too slow
    @pytest.mark.skipif(**tm.no_sklearn())
    def test_pickling(self) -> None:
        x, y = build_dataset()
        train_x = xgb.DMatrix(x, label=y)

        param = {"tree_method": "hist", "device": "cuda"}
        bst = xgb.train(param, train_x)
        self.run_pickling(bst)

        bst = xgb.XGBRegressor(**param).fit(x, y)
        self.run_pickling(bst)

        param = {"booster": "gblinear", "updater": "coord_descent", "device": "cuda"}
        bst = xgb.train(param, train_x)
        self.run_pickling(bst)

        bst = xgb.XGBRegressor(**param).fit(x, y)
        self.run_pickling(bst)

    @pytest.mark.mgpu
    def test_wrap_gpu_id(self) -> None:
        X, y = build_dataset()
        dtrain = xgb.DMatrix(X, y)

        bst = xgb.train(
            {"tree_method": "hist", "device": "cuda:1"}, dtrain, num_boost_round=6
        )

        model_path = "model.pkl"
        save_pickle(bst, model_path)
        cuda_environment = {"CUDA_VISIBLE_DEVICES": "0"}
        env = os.environ.copy()
        env.update(cuda_environment)
        args = self.args_template.copy()
        args.append(
            "./tests/python-gpu/" "load_pickle.py::TestLoadPickle::test_wrap_gpu_id"
        )
        status = subprocess.call(args, env=env)
        assert status == 0
        os.remove(model_path)

    def test_pickled_context(self) -> None:
        x, y = tm.make_sparse_regression(10, 10, sparsity=0.8, as_dense=True)
        train_x = xgb.DMatrix(x, label=y)

        def run_test(param: Dict[str, Any]) -> None:
            bst = xgb.train(param, train_x)

            save_pickle(bst, model_path)

            args = self.args_template.copy()
            root = tm.project_root(__file__)
            path = os.path.join(root, "tests", "python-gpu", "load_pickle.py")
            args.append(path + "::TestLoadPickle::test_context_is_removed")

            cuda_environment = {"CUDA_VISIBLE_DEVICES": "-1"}
            env = os.environ.copy()
            env.update(cuda_environment)

            # Load model in a CPU only environment.
            status = subprocess.call(args, env=env)
            assert status == 0

            args = self.args_template.copy()
            args.append(
                "./tests/python-gpu/"
                "load_pickle.py::TestLoadPickle::test_context_is_preserved"
            )

            # Load in environment that has GPU.
            env = os.environ.copy()
            assert "CUDA_VISIBLE_DEVICES" not in env.keys()
            status = subprocess.call(args, env=env)
            assert status == 0

            os.remove(model_path)

        param = {"tree_method": "hist", "verbosity": 1, "device": "cuda"}
        run_test(param)
        param = {"booster": "gblinear", "updater": "coord_descent", "device": "cuda"}
        run_test(param)

    @pytest.mark.skipif(**tm.no_sklearn())
    def test_predict_sklearn_pickle(self) -> None:
        from sklearn.datasets import load_digits

        x, y = load_digits(return_X_y=True)

        kwargs = {
            "tree_method": "hist",
            "objective": "binary:logistic",
            "device": "cuda",
            "n_estimators": 10,
        }

        model = XGBClassifier(**kwargs)
        model.fit(x, y)

        save_pickle(model, "model.pkl")
        del model

        # load model
        model = load_pickle("model.pkl")
        os.remove("model.pkl")

        gpu_pred = model.predict(x, output_margin=True)

        # Switch to CPU predictor
        bst = model.get_booster()
        bst.set_param({"device": "cpu"})
        cpu_pred = model.predict(x, output_margin=True)
        np.testing.assert_allclose(cpu_pred, gpu_pred, rtol=1e-5)

    def test_training_on_cpu_only_env(self) -> None:
        cuda_environment = {"CUDA_VISIBLE_DEVICES": "-1"}
        env = os.environ.copy()
        env.update(cuda_environment)
        args = self.args_template.copy()
        args.append(
            "./tests/python-gpu/"
            "load_pickle.py::TestLoadPickle::test_training_on_cpu_only_env"
        )
        status = subprocess.call(args, env=env)
        assert status == 0



"""
Demonstration for parsing JSON/UBJSON tree model files
======================================================

See :doc:`/tutorials/saving_model` for details about the model serialization.

"""

import argparse
import json
from dataclasses import dataclass
from enum import IntEnum, unique
from typing import Any, Dict, List, Sequence, Union

import numpy as np

try:
    import ubjson
except ImportError:
    ubjson = None


ParamT = Dict[str, str]


def to_integers(data: Union[bytes, List[int]]) -> List[int]:
    """Convert a sequence of bytes to a list of Python integer"""
    return [v for v in data]


@unique
class SplitType(IntEnum):
    numerical = 0
    categorical = 1


@dataclass
class Node:
    # properties
    left: int
    right: int
    parent: int
    split_idx: int
    split_cond: float
    default_left: bool
    split_type: SplitType
    categories: List[int]
    # statistic
    base_weight: float
    loss_chg: float
    sum_hess: float


class Tree:
    """A tree built by XGBoost."""

    def __init__(self, tree_id: int, nodes: Sequence[Node]) -> None:
        self.tree_id = tree_id
        self.nodes = nodes

    def loss_change(self, node_id: int) -> float:
        """Loss gain of a node."""
        return self.nodes[node_id].loss_chg

    def sum_hessian(self, node_id: int) -> float:
        """Sum Hessian of a node."""
        return self.nodes[node_id].sum_hess

    def base_weight(self, node_id: int) -> float:
        """Base weight of a node."""
        return self.nodes[node_id].base_weight

    def split_index(self, node_id: int) -> int:
        """Split feature index of node."""
        return self.nodes[node_id].split_idx

    def split_condition(self, node_id: int) -> float:
        """Split value of a node."""
        return self.nodes[node_id].split_cond

    def split_categories(self, node_id: int) -> List[int]:
        """Categories in a node."""
        return self.nodes[node_id].categories

    def is_categorical(self, node_id: int) -> bool:
        """Whether a node has categorical split."""
        return self.nodes[node_id].split_type == SplitType.categorical

    def is_numerical(self, node_id: int) -> bool:
        return not self.is_categorical(node_id)

    def parent(self, node_id: int) -> int:
        """Parent ID of a node."""
        return self.nodes[node_id].parent

    def left_child(self, node_id: int) -> int:
        """Left child ID of a node."""
        return self.nodes[node_id].left

    def right_child(self, node_id: int) -> int:
        """Right child ID of a node."""
        return self.nodes[node_id].right

    def is_leaf(self, node_id: int) -> bool:
        """Whether a node is leaf."""
        return self.nodes[node_id].left == -1

    def is_deleted(self, node_id: int) -> bool:
        """Whether a node is deleted."""
        return self.split_index(node_id) == np.iinfo(np.uint32).max

    def __str__(self) -> str:
        stack = [0]
        nodes = []
        while stack:
            node: Dict[str, Union[float, int, List[int]]] = {}
            nid = stack.pop()

            node["node id"] = nid
            node["gain"] = self.loss_change(nid)
            node["cover"] = self.sum_hessian(nid)
            nodes.append(node)

            if not self.is_leaf(nid) and not self.is_deleted(nid):
                left = self.left_child(nid)
                right = self.right_child(nid)
                stack.append(left)
                stack.append(right)
                categories = self.split_categories(nid)
                if categories:
                    assert self.is_categorical(nid)
                    node["categories"] = categories
                else:
                    assert self.is_numerical(nid)
                    node["condition"] = self.split_condition(nid)
            if self.is_leaf(nid):
                node["weight"] = self.split_condition(nid)

        string = "\n".join(map(lambda x: "  " + str(x), nodes))
        return string


class Model:
    """Gradient boosted tree model."""

    def __init__(self, model: dict) -> None:
        """Construct the Model from a JSON object.

        parameters
        ----------
         model : A dictionary loaded by json representing a XGBoost boosted tree model.
        """
        # Basic properties of a model
        self.learner_model_shape: ParamT = model["learner"]["learner_model_param"]
        self.num_output_group = int(self.learner_model_shape["num_class"])
        self.num_feature = int(self.learner_model_shape["num_feature"])
        self.base_score: List[float] = json.loads(
            self.learner_model_shape["base_score"]
        )
        # A field encoding which output group a tree belongs
        self.tree_info = model["learner"]["gradient_booster"]["model"]["tree_info"]

        model_shape: ParamT = model["learner"]["gradient_booster"]["model"][
            "gbtree_model_param"
        ]

        # JSON representation of trees
        j_trees = model["learner"]["gradient_booster"]["model"]["trees"]

        # Load the trees
        self.num_trees = int(model_shape["num_trees"])

        trees: List[Tree] = []
        for i in range(self.num_trees):
            tree: Dict[str, Any] = j_trees[i]
            tree_id = int(tree["id"])
            assert tree_id == i, (tree_id, i)
            # - properties
            left_children: List[int] = tree["left_children"]
            right_children: List[int] = tree["right_children"]
            parents: List[int] = tree["parents"]
            split_conditions: List[float] = tree["split_conditions"]
            split_indices: List[int] = tree["split_indices"]
            # when ubjson is used, this is a byte array with each element as uint8
            default_left = to_integers(tree["default_left"])

            # - categorical features
            # when ubjson is used, this is a byte array with each element as uint8
            split_types = to_integers(tree["split_type"])
            # categories for each node is stored in a CSR style storage with segment as
            # the begin ptr and the `categories' as values.
            cat_segments: List[int] = tree["categories_segments"]
            cat_sizes: List[int] = tree["categories_sizes"]
            # node index for categorical nodes
            cat_nodes: List[int] = tree["categories_nodes"]
            assert len(cat_segments) == len(cat_sizes) == len(cat_nodes)
            cats = tree["categories"]
            assert len(left_children) == len(split_types)

            # The storage for categories is only defined for categorical nodes to
            # prevent unnecessary overhead for numerical splits, we track the
            # categorical node that are processed using a counter.
            cat_cnt = 0
            if cat_nodes:
                last_cat_node = cat_nodes[cat_cnt]
            else:
                last_cat_node = -1
            node_categories: List[List[int]] = []
            for node_id in range(len(left_children)):
                if node_id == last_cat_node:
                    beg = cat_segments[cat_cnt]
                    size = cat_sizes[cat_cnt]
                    end = beg + size
                    node_cats = cats[beg:end]
                    # categories are unique for each node
                    assert len(set(node_cats)) == len(node_cats)
                    cat_cnt += 1
                    if cat_cnt == len(cat_nodes):
                        last_cat_node = -1  # continue to process the rest of the nodes
                    else:
                        last_cat_node = cat_nodes[cat_cnt]
                    assert node_cats
                    node_categories.append(node_cats)
                else:
                    # append an empty node, it's either a numerical node or a leaf.
                    node_categories.append([])

            # - stats
            base_weights: List[float] = tree["base_weights"]
            loss_changes: List[float] = tree["loss_changes"]
            sum_hessian: List[float] = tree["sum_hessian"]

            # Construct a list of nodes that have complete information
            nodes: List[Node] = [
                Node(
                    left_children[node_id],
                    right_children[node_id],
                    parents[node_id],
                    split_indices[node_id],
                    split_conditions[node_id],
                    default_left[node_id] == 1,  # to boolean
                    SplitType(split_types[node_id]),
                    node_categories[node_id],
                    base_weights[node_id],
                    loss_changes[node_id],
                    sum_hessian[node_id],
                )
                for node_id in range(len(left_children))
            ]

            pytree = Tree(tree_id, nodes)
            trees.append(pytree)

        self.trees = trees

    def print_model(self) -> None:
        for i, tree in enumerate(self.trees):
            print("\ntree_id:", i)
            print(tree)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Demonstration for loading XGBoost JSON/UBJSON model."
    )
    parser.add_argument(
        "--model", type=str, required=True, help="Path to .json/.ubj model file."
    )
    args = parser.parse_args()
    if args.model.endswith("json"):
        # use json format
        with open(args.model, "r") as fd:
            model = json.load(fd)
    elif args.model.endswith("ubj"):
        if ubjson is None:
            raise ImportError("ubjson is not installed.")
        # use ubjson format
        with open(args.model, "rb") as bfd:
            model = ubjson.load(bfd)
    else:
        raise ValueError(
            "Unexpected file extension. Supported file extension are json and ubj."
        )
    model = Model(model)
    model.print_model()



from typing import Any, Dict

import numpy as np
import pytest
from hypothesis import assume, given, note, settings, strategies

import xgboost as xgb
from xgboost import testing as tm
from xgboost.testing.params import (
    cat_parameter_strategy,
    exact_parameter_strategy,
    hist_cache_strategy,
    hist_parameter_strategy,
)
from xgboost.testing.updater import (
    check_categorical_missing,
    check_categorical_ohe,
    check_get_quantile_cut,
    check_quantile_loss,
    run_invalid_category,
    run_max_cat,
    train_result,
)

pytestmark = tm.timeout(30)


class TestGPUUpdaters:
    @given(
        exact_parameter_strategy,
        hist_parameter_strategy,
        hist_cache_strategy,
        strategies.integers(1, 20),
        tm.make_dataset_strategy(),
    )
    @settings(deadline=None, max_examples=50, print_blob=True)
    def test_gpu_hist(
        self,
        param: Dict[str, Any],
        hist_param: Dict[str, Any],
        cache_param: Dict[str, Any],
        num_rounds: int,
        dataset: tm.TestDataset,
    ) -> None:
        param.update({"tree_method": "hist", "device": "cuda"})
        param.update(hist_param)
        param.update(cache_param)
        param = dataset.set_params(param)
        result = train_result(param, dataset.get_dmat(), num_rounds)
        note(str(result))
        assert tm.non_increasing(result["train"][dataset.metric])

    @pytest.mark.parametrize("tree_method", ["approx", "hist"])
    def test_cache_size(self, tree_method: str) -> None:
        from sklearn.datasets import make_regression

        X, y = make_regression(n_samples=4096, n_features=64, random_state=1994)
        Xy = xgb.DMatrix(X, y)
        results = []
        for cache_size in [1, 3, 2048]:
            params: Dict[str, Any] = {"tree_method": tree_method, "device": "cuda"}
            params["max_cached_hist_node"] = cache_size
            evals_result: Dict[str, Dict[str, list]] = {}
            xgb.train(
                params,
                Xy,
                num_boost_round=4,
                evals=[(Xy, "Train")],
                evals_result=evals_result,
            )
            results.append(evals_result["Train"]["rmse"])
        for i in range(1, len(results)):
            np.testing.assert_allclose(results[0], results[i])

    @given(
        exact_parameter_strategy,
        hist_parameter_strategy,
        hist_cache_strategy,
        strategies.integers(1, 20),
        tm.make_dataset_strategy(),
    )
    @settings(deadline=None, print_blob=True)
    def test_gpu_approx(
        self,
        param: Dict[str, Any],
        hist_param: Dict[str, Any],
        cache_param: Dict[str, Any],
        num_rounds: int,
        dataset: tm.TestDataset,
    ) -> None:
        param.update({"tree_method": "approx", "device": "cuda"})
        param.update(hist_param)
        param.update(cache_param)
        param = dataset.set_params(param)
        result = train_result(param, dataset.get_dmat(), num_rounds)
        note(str(result))
        assert tm.non_increasing(result["train"][dataset.metric])

    @given(tm.sparse_datasets_strategy)
    @settings(deadline=None, print_blob=True)
    def test_sparse(self, dataset: tm.TestDataset) -> None:
        param = {"tree_method": "hist", "max_bin": 64}
        hist_result = train_result(param, dataset.get_dmat(), 16)
        note(str(hist_result))
        assert tm.non_increasing(hist_result["train"][dataset.metric])

        param = {"tree_method": "hist", "max_bin": 64, "device": "cuda"}
        gpu_hist_result = train_result(param, dataset.get_dmat(), 16)
        note(str(gpu_hist_result))
        assert tm.non_increasing(gpu_hist_result["train"][dataset.metric])

        np.testing.assert_allclose(
            hist_result["train"]["rmse"], gpu_hist_result["train"]["rmse"], rtol=1e-2
        )

    @given(
        strategies.integers(10, 400),
        strategies.integers(3, 8),
        strategies.integers(1, 2),
        strategies.integers(4, 7),
    )
    @settings(deadline=None, max_examples=20, print_blob=True)
    @pytest.mark.skipif(**tm.no_pandas())
    def test_categorical_ohe(
        self, rows: int, cols: int, rounds: int, cats: int
    ) -> None:
        check_categorical_ohe(
            rows=rows,
            cols=cols,
            rounds=rounds,
            cats=cats,
            device="cuda",
            tree_method="hist",
            extmem=False,
        )

    @given(
        tm.categorical_dataset_strategy,
        hist_parameter_strategy,
        cat_parameter_strategy,
        strategies.integers(4, 32),
    )
    @settings(deadline=None, max_examples=20, print_blob=True)
    @pytest.mark.skipif(**tm.no_pandas())
    def test_categorical_hist(
        self,
        dataset: tm.TestDataset,
        hist_parameters: Dict[str, Any],
        cat_parameters: Dict[str, Any],
        n_rounds: int,
    ) -> None:
        cat_parameters.update(hist_parameters)
        cat_parameters["tree_method"] = "hist"
        cat_parameters["device"] = "cuda"

        results = train_result(cat_parameters, dataset.get_dmat(), n_rounds)
        tm.non_increasing(results["train"]["rmse"])

    @given(
        tm.categorical_dataset_strategy,
        hist_parameter_strategy,
        cat_parameter_strategy,
        strategies.integers(4, 32),
    )
    @settings(deadline=None, max_examples=20, print_blob=True)
    @pytest.mark.skipif(**tm.no_pandas())
    def test_categorical_approx(
        self,
        dataset: tm.TestDataset,
        hist_parameters: Dict[str, Any],
        cat_parameters: Dict[str, Any],
        n_rounds: int,
    ) -> None:
        cat_parameters.update(hist_parameters)
        cat_parameters["tree_method"] = "approx"
        cat_parameters["device"] = "cuda"

        results = train_result(cat_parameters, dataset.get_dmat(), n_rounds)
        tm.non_increasing(results["train"]["rmse"])

    @given(
        hist_parameter_strategy,
        cat_parameter_strategy,
    )
    @settings(deadline=None, max_examples=10, print_blob=True)
    def test_categorical_ames_housing(
        self,
        hist_parameters: Dict[str, Any],
        cat_parameters: Dict[str, Any],
    ) -> None:
        cat_parameters.update(hist_parameters)
        dataset = tm.TestDataset(
            "ames_housing", tm.data.get_ames_housing, "reg:squarederror", "rmse"
        )
        cat_parameters["tree_method"] = "hist"
        cat_parameters["device"] = "cuda"
        results = train_result(cat_parameters, dataset.get_dmat(), 16)
        tm.non_increasing(results["train"]["rmse"])

    @given(
        strategies.integers(10, 400),
        strategies.integers(3, 8),
        strategies.integers(4, 7),
    )
    @settings(deadline=None, max_examples=20, print_blob=True)
    @pytest.mark.skipif(**tm.no_pandas())
    def test_categorical_missing(self, rows: int, cols: int, cats: int) -> None:
        check_categorical_missing(
            rows, cols, cats, device="cuda", tree_method="approx", extmem=False
        )
        check_categorical_missing(
            rows, cols, cats, device="cuda", tree_method="hist", extmem=False
        )

    @pytest.mark.skipif(**tm.no_pandas())
    @pytest.mark.parametrize("tree_method", ["hist", "approx"])
    def test_max_cat(self, tree_method: str) -> None:
        run_max_cat(tree_method, "cuda")

    def test_categorical_32_cat(self) -> None:
        """32 hits the bound of integer bitset, so special test"""
        rows = 1000
        check_categorical_ohe(
            rows=rows,
            cols=10,
            rounds=4,
            cats=32,
            device="cuda",
            tree_method="hist",
            extmem=False,
        )

    @pytest.mark.skipif(**tm.no_cupy())
    @pytest.mark.parametrize("tree_method", ["hist", "approx"])
    def test_invalid_category(self, tree_method: str) -> None:
        run_invalid_category(tree_method, "cuda")

    @pytest.mark.skipif(**tm.no_cupy())
    @given(
        hist_parameter_strategy,
        strategies.integers(1, 20),
        tm.make_dataset_strategy(),
    )
    @settings(deadline=None, max_examples=20, print_blob=True)
    def test_gpu_hist_device_dmatrix(
        self, param: dict, num_rounds: int, dataset: tm.TestDataset
    ) -> None:
        # We cannot handle empty dataset yet
        assume(len(dataset.y) > 0)
        param["tree_method"] = "hist"
        param["device"] = "cuda"
        param = dataset.set_params(param)
        result = train_result(
            param,
            dataset.get_device_dmat(max_bin=param.get("max_bin", None)),
            num_rounds,
        )
        note(str(result))
        assert tm.non_increasing(result["train"][dataset.metric], tolerance=1e-3)

    @given(
        hist_parameter_strategy,
        strategies.integers(1, 3),
        tm.make_dataset_strategy(),
    )
    @settings(deadline=None, max_examples=10, print_blob=True)
    def test_external_memory(
        self, param: Dict[str, Any], num_rounds: int, dataset: tm.TestDataset
    ) -> None:
        # We cannot handle empty dataset yet
        assume(len(dataset.y) > 0)

        with xgb.config_context(use_rmm=True):
            param["tree_method"] = "hist"
            param["device"] = "cuda"
            param = dataset.set_params(param)
            m = dataset.get_external_dmat()
            external_result = train_result(param, m, num_rounds)
            del m
            assert tm.non_increasing(external_result["train"][dataset.metric])

    def test_empty_dmatrix_prediction(self) -> None:
        # FIXME(trivialfis): This should be done with all updaters
        kRows = 0
        kCols = 100

        X = np.empty((kRows, kCols))
        y = np.empty((kRows,))

        dtrain = xgb.DMatrix(X, y)

        bst = xgb.train(
            {"verbosity": 2, "tree_method": "hist", "device": "cuda"},
            dtrain,
            verbose_eval=True,
            num_boost_round=6,
            evals=[(dtrain, "Train")],
        )

        kRows = 100
        X_test = np.random.randn(kRows, kCols)

        dtest = xgb.DMatrix(X_test)
        predictions = bst.predict(dtest)
        # non-distributed, 0.0 is returned due to base_score estimation with 0 gradient.
        np.testing.assert_allclose(predictions, 0.0, 1e-6)

    @pytest.mark.mgpu
    @given(tm.make_dataset_strategy(), strategies.integers(0, 10))
    @settings(deadline=None, max_examples=10, print_blob=True)
    def test_specified_gpu_id_gpu_update(
        self, dataset: tm.TestDataset, gpu_id: int
    ) -> None:
        param = {"tree_method": "hist", "device": f"cuda:{gpu_id}"}
        param = dataset.set_params(param)
        result = train_result(param, dataset.get_dmat(), 10)
        assert tm.non_increasing(result["train"][dataset.metric])

    @pytest.mark.parametrize("weighted", [True, False])
    def test_quantile_loss(self, weighted: bool) -> None:
        check_quantile_loss("hist", weighted, "cuda")

    @pytest.mark.skipif(**tm.no_pandas())
    def test_issue8824(self) -> None:
        # column sampling by node crashes because shared pointers go out of scope
        import pandas as pd

        data = pd.DataFrame(np.random.rand(1024, 8))
        data.columns = "x" + data.columns.astype(str)
        features = data.columns
        data["y"] = data.sum(axis=1) < 4
        dtrain = xgb.DMatrix(data[features], label=data["y"])
        model = xgb.train(
            dtrain=dtrain,
            params={
                "max_depth": 5,
                "learning_rate": 0.05,
                "objective": "binary:logistic",
                "tree_method": "hist",
                "device": "cuda",
                "colsample_bytree": 0.5,
                "colsample_bylevel": 0.5,
                "colsample_bynode": 0.5,  # Causes issues
                "reg_alpha": 0.05,
                "reg_lambda": 0.005,
                "seed": 66,
                "subsample": 0.5,
                "gamma": 0.2,
                "eval_metric": "auc",
            },
            num_boost_round=150,
        )

    @pytest.mark.skipif(**tm.no_cudf())
    def test_get_quantile_cut(self) -> None:
        check_get_quantile_cut("hist", "cuda")



"""Helper script for triggering Read the docs build.

See `doc/contrib/docs.rst <https://xgboost.readthedocs.io/en/stable/contrib/docs.html>`__
for more info.

"""

import json
import os
import pprint
from http.client import responses as http_responses

import requests  # type: ignore


def trigger_build(token: str) -> None:
    """Trigger RTD build."""

    event_path = os.environ["GITHUB_EVENT_PATH"]
    with open(event_path, "r") as fd:
        event: dict = json.load(fd)

    if event.get("pull_request", None) is None:
        # refs/heads/branch-name
        branch = event["ref"].split("/")[-1]
    else:
        branch = event["pull_request"]["number"]

    if branch == "master":
        # Use the `latest` tag, otherwise RTD wouldn't update the rendered doc.
        branch = "latest"

    URL = f"https://readthedocs.org/api/v3/projects/xgboost/versions/{branch}/builds/"
    HEADERS = {"Authorization": f"token {token}"}
    response = requests.post(URL, headers=HEADERS)
    # 202 means the build is successfully triggered.
    if response.status_code != 202:
        status_text = http_responses[response.status_code]
        raise RuntimeError(
            "ReadTheDocs returned an unexpected response: "
            f"{response.status_code} {status_text}, reason: {response.reason}"
        )
    pprint.pprint(response.json(), indent=4)


def main() -> None:
    token = os.getenv("RTD_AUTH_TOKEN")
    # GA redacts the secret by default, but we should still be really careful to not log
    # (expose) the token in the CI.
    if token is None:
        raise RuntimeError(
            "The RTD_AUTH_TOKEN environment variable must be set to a valid auth token for the"
            "ReadTheDocs service."
        )
    if len(token) == 0:
        print("Document build is not triggered.")
        return

    if not isinstance(token, str) or len(token) != 40:
        raise ValueError(f"Invalid token.")

    trigger_build(token)


if __name__ == "__main__":
    main()



"""
Getting started with categorical data
=====================================

Experimental support for categorical data.

In before, users need to run an encoder themselves before passing the data into XGBoost,
which creates a sparse matrix and potentially increase memory usage.  This demo
showcases the experimental categorical data support, more advanced features are planned.

  .. versionadded:: 1.5.0

See Also
--------
- :doc:`Tutorial </tutorials/categorical>`
- :ref:`sphx_glr_python_examples_cat_in_the_dat.py`
- :ref:`sphx_glr_python_examples_cat_pipeline.py`

"""

from typing import Tuple

import numpy as np
import pandas as pd

import xgboost as xgb


def make_categorical(
    n_samples: int, n_features: int, n_categories: int, onehot: bool
) -> Tuple[pd.DataFrame, pd.Series]:
    """Make some random data for demo."""
    rng = np.random.RandomState(1994)

    pd_dict = {}
    for i in range(n_features + 1):
        c = rng.randint(low=0, high=n_categories, size=n_samples)
        pd_dict[str(i)] = pd.Series(c, dtype=np.int64)

    df = pd.DataFrame(pd_dict)
    label = df.iloc[:, 0]
    df = df.iloc[:, 1:]
    for i in range(0, n_features):
        label += df.iloc[:, i]
    label += 1

    df = df.astype("category")
    categories = np.arange(0, n_categories)
    for col in df.columns:
        df[col] = df[col].cat.set_categories(categories)

    if onehot:
        return pd.get_dummies(df), label
    return df, label


def main() -> None:
    # Use builtin categorical data support

    # For scikit-learn interface, the input data should be pandas DataFrame or cudf
    # DataFrame with categorical features. If an numpy/cupy array is used instead, the
    # `feature_types` for `XGBRegressor` should be set accordingly.
    X, y = make_categorical(100, 10, 4, False)
    # Specify `enable_categorical` to True, also we use onehot-encoding-based split here
    # for demonstration. For details see the document of `max_cat_to_onehot`.
    reg = xgb.XGBRegressor(
        tree_method="hist", enable_categorical=True, max_cat_to_onehot=5, device="cuda"
    )
    reg.fit(X, y, eval_set=[(X, y)])

    # Pass in already encoded data
    X_enc, y_enc = make_categorical(100, 10, 4, True)
    reg_enc = xgb.XGBRegressor(tree_method="hist", device="cuda")
    reg_enc.fit(X_enc, y_enc, eval_set=[(X_enc, y_enc)])

    reg_results = np.array(reg.evals_result()["validation_0"]["rmse"])
    reg_enc_results = np.array(reg_enc.evals_result()["validation_0"]["rmse"])

    # Check that they have same results
    np.testing.assert_allclose(reg_results, reg_enc_results)

    # Convert to DMatrix for SHAP value
    booster: xgb.Booster = reg.get_booster()
    m = xgb.DMatrix(X, enable_categorical=True)  # specify categorical data support.
    SHAP = booster.predict(m, pred_contribs=True)
    margin = booster.predict(m, output_margin=True)
    np.testing.assert_allclose(
        np.sum(SHAP, axis=len(SHAP.shape) - 1), margin, rtol=1e-3
    )


if __name__ == "__main__":
    main()



"""
Helper script to prepare for releasing XGBoost JVM packages to
Maven Central.

## Prerequisite

1. You must have the right to upload artifacts to the Maven Central repo.
   If you do not, contact Hyunsu Cho (chohyu01@cs.washington.edu) so that
   he can contact Sonatype on your behalf in order to add you as a
   "producer" user for the ml.dmlc namespace. See
   https://central.sonatype.org/pages/support/#status to learn about
   the process for adding or removing users who can publish to the project.

2. Follow instructions in
   https://central.sonatype.org/publish/publish-portal-maven/#credentials
   to set up the authentication token in your machine.

3. Set up GPG for signing artifacts:
   https://central.sonatype.org/publish/requirements/gpg/

## Making the release
Run this script 4 times:

python3 dev/prepare_jvm_release.py --scala-version 2.12 --variant cpu
python3 dev/prepare_jvm_release.py --scala-version 2.12 --variant gpu
python3 dev/prepare_jvm_release.py --scala-version 2.13 --variant cpu
python3 dev/prepare_jvm_release.py --scala-version 2.13 --variant gpu

"""

import argparse
import errno
import glob
import os
import re
import shutil
import subprocess
import sys
import tempfile
import zipfile
from contextlib import contextmanager
from urllib.request import urlretrieve


def normpath(path):
    """Normalize UNIX path to a native path."""
    normalized = os.path.join(*path.split("/"))
    if os.path.isabs(path):
        return os.path.abspath("/") + normalized
    else:
        return normalized


def cp(source, target):
    source = normpath(source)
    target = normpath(target)
    print("cp {0} {1}".format(source, target))
    shutil.copy(source, target)


def maybe_makedirs(path):
    path = normpath(path)
    print("mkdir -p " + path)
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


@contextmanager
def cd(path):
    path = normpath(path)
    cwd = os.getcwd()
    os.chdir(path)
    print("cd " + path)
    try:
        yield path
    finally:
        os.chdir(cwd)


def run(command, **kwargs):
    print(command)
    subprocess.run(command, shell=True, check=True, **kwargs)


def get_current_commit_hash():
    out = subprocess.check_output(["git", "rev-parse", "HEAD"])
    return out.decode().split("\n")[0]


def get_current_git_branch():
    out = subprocess.check_output(["git", "log", "-n", "1", "--pretty=%d", "HEAD"])
    m = re.search(r"release_[0-9\.]+", out.decode())
    if not m:
        raise ValueError("Expected branch name of form release_xxx")
    return m.group(0)


def retrieve(url, filename=None):
    print(f"{url} -> {filename}")
    return urlretrieve(url, filename)


def main():
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawTextHelpFormatter
    )
    parser.add_argument(
        "--release-version",
        type=str,
        required=True,
        help="Version of the release being prepared",
    )
    parser.add_argument(
        "--scala-version",
        type=str,
        required=True,
        help="Version of Scala to use in the JVM packages",
        choices=["2.12", "2.13"],
    )
    parser.add_argument(
        "--variant",
        type=str,
        required=True,
        choices=["cpu", "gpu"],
        help="JVM package variant to package and publish",
    )

    args = parser.parse_args()
    version = args.release_version
    scala_version = args.scala_version
    use_cuda = args.variant == "gpu"

    commit_hash = get_current_commit_hash()
    git_branch = get_current_git_branch()
    print(f"Using commit {commit_hash} of branch {git_branch}")
    print(f"====Update pom.xml to use Scala {scala_version}====")
    run(
        f"{sys.executable} ops/script/change_scala_version.py "
        f"--scala-version {scala_version} --purge-artifacts"
    )

    with cd("jvm-packages/"):
        print("====Copying resources for testing====")
        with cd("../demo/data/regression"):
            run(f"{sys.executable} mapfeat.py")
            run(f"{sys.executable} mknfold.py machine.txt 1")
        xgboost4j_spark = "xgboost4j-spark-gpu" if use_cuda else "xgboost4j-spark"
        maybe_makedirs("xgboost4j/src/test/resources")
        maybe_makedirs(f"{xgboost4j_spark}/src/test/resources")
        for file in glob.glob("../demo/data/agaricus.*"):
            cp(file, "xgboost4j/src/test/resources")
            cp(file, f"{xgboost4j_spark}/src/test/resources")
        for file in glob.glob("../demo/data/regression/machine.txt.t*"):
            cp(file, f"{xgboost4j_spark}/src/test/resources")

        print("====Creating directories to hold native binaries====")
        if use_cuda:
            # TODO(hcho3): Add GPU build for linux aarch64
            matrix = [("linux", "x86_64")]
        else:
            matrix = [
                ("linux", "x86_64"),
                ("linux", "aarch64"),
                ("windows", "x86_64"),
                ("macos", "x86_64"),
                ("macos", "aarch64"),
            ]
        for os_ident, arch in matrix:
            output_dir = f"xgboost4j/src/main/resources/lib/{os_ident}/{arch}"
            maybe_makedirs(output_dir)

        print("====Downloading native binaries from CI====")
        if use_cuda:
            url_prefix = (
                "https://s3-us-west-2.amazonaws.com/xgboost-maven-repo/release/ml/dmlc"
            )
            with tempfile.TemporaryDirectory() as tempdir:
                # libxgboost4j.so for Linux x86_64, GPU support
                zip_path = os.path.join(tempdir, "xgboost4j-spark-gpu_2.12.jar")
                extract_dir = os.path.join(tempdir, "xgboost4j-spark-gpu")
                retrieve(
                    url=f"{url_prefix}/xgboost4j-spark-gpu_2.12/{version}/"
                    f"xgboost4j-spark-gpu_2.12-{version}.jar",
                    filename=zip_path,
                )
                os.mkdir(extract_dir)
                with zipfile.ZipFile(zip_path, "r") as t:
                    t.extractall(extract_dir)
                cp(
                    os.path.join(
                        extract_dir, "lib", "linux", "x86_64", "libxgboost4j.so"
                    ),
                    "xgboost4j/src/main/resources/lib/linux/x86_64/libxgboost4j.so",
                )
            run(
                "mvn --no-transfer-progress install -Pgpu "
                "-DskipTests -Dmaven.test.skip=true -Dskip.native.build=true"
            )
            run(
                "mvn deploy -Pgpu,release -pl xgboost4j-spark-gpu "
                "-DskipTests -Dmaven.test.skip=true -Dskip.native.build=true"
            )
        else:
            url_prefix = "https://s3-us-west-2.amazonaws.com/xgboost-nightly-builds"
            for os_ident, arch, src_libname, dest_libname in [
                ("linux", "x86_64", "libxgboost4j_linux_x86_64.so", "libxgboost4j.so"),
                (
                    "linux",
                    "aarch64",
                    "libxgboost4j_linux_aarch64.so",
                    "libxgboost4j.so",
                ),
                ("windows", "x86_64", "xgboost4j.dll", "xgboost4j.dll"),
                ("macos", "x86_64", "libxgboost4j_intel.dylib", "libxgboost4j.dylib"),
                ("macos", "aarch64", "libxgboost4j_m1.dylib", "libxgboost4j.dylib"),
            ]:
                retrieve(
                    url=f"{url_prefix}/{git_branch}/{commit_hash}/{src_libname}",
                    filename=(
                        "xgboost4j/src/main/resources/lib/"
                        f"{os_ident}/{arch}/{dest_libname}"
                    ),
                )
            run(
                "mvn --no-transfer-progress deploy -Pdefault,release "
                "-DskipTests -Dmaven.test.skip=true -Dskip.native.build=true"
            )

    print("====Next Steps====")
    print(
        "Visit https://central.sonatype.com/publishing/deployments to verify the deployment. "
        "You can either drop the deployment or publish it. Note: publishing is final."
    )


if __name__ == "__main__":
    main()



import pytest

from xgboost import testing as tm
from xgboost.testing.interaction_constraints import (
    run_interaction_constraints,
    training_accuracy,
)


class TestInteractionConstraints:
    def test_exact_interaction_constraints(self) -> None:
        run_interaction_constraints(tree_method="exact", device="cpu")

    def test_hist_interaction_constraints(self) -> None:
        run_interaction_constraints(tree_method="hist", device="cpu")

    def test_approx_interaction_constraints(self) -> None:
        run_interaction_constraints(tree_method="approx", device="cpu")

    def test_interaction_constraints_feature_names(self) -> None:
        with pytest.raises(ValueError):
            constraints = [("feature_0", "feature_1")]
            run_interaction_constraints(
                tree_method="exact", device="cpu", interaction_constraints=constraints
            )

        with pytest.raises(ValueError):
            constraints = [("feature_0", "feature_3")]
            feature_names = ["feature_0", "feature_1", "feature_2"]
            run_interaction_constraints(
                tree_method="exact",
                device="cpu",
                feature_names=feature_names,
                interaction_constraints=constraints,
            )

        constraints = [("feature_0", "feature_1")]
        feature_names = ["feature_0", "feature_1", "feature_2"]
        run_interaction_constraints(
            tree_method="exact",
            device="cpu",
            feature_names=feature_names,
            interaction_constraints=constraints,
        )

        constraints_lst = [["feature_0", "feature_1"], ["feature_2"]]
        feature_names = ["feature_0", "feature_1", "feature_2"]
        run_interaction_constraints(
            tree_method="exact",
            device="cpu",
            feature_names=feature_names,
            interaction_constraints=constraints_lst,
        )

    @pytest.mark.skipif(**tm.no_sklearn())
    @pytest.mark.parametrize("tree_method", ["hist", "approx", "exact"])
    def test_hist_training_accuracy(self, tree_method: str) -> None:
        dpath = "demo/data/"
        training_accuracy(tree_method=tree_method, dpath=dpath, device="cpu")



#!/usr/bin/env python
"""Build the native XGBoost4J JNI library."""

import argparse
import os
import platform
import shlex
import shutil
import subprocess
import sys
from pathlib import Path
from typing import Sequence

ROOT = Path(__file__).resolve().parents[1]
JVM_PACKAGES = Path(__file__).resolve().parent

DEFAULT_CONFIG = {
    "USE_OPENMP": "ON",
    "USE_CUDA": "OFF",
    "USE_NCCL": "OFF",
    "JVM_BINDINGS": "ON",
    "LOG_CAPI_INVOCATION": "OFF",
    "CMAKE_EXPORT_COMPILE_COMMANDS": "ON",
}


def run(command: Sequence[str], *, cwd: Path | None = None) -> None:
    """Run a shell command."""
    print(shlex.join(command), flush=True)
    subprocess.run(command, cwd=cwd, check=True, env=os.environ)


def mkdir(path: Path) -> None:
    """Create a directory if it does not already exist."""
    print(f"mkdir -p {path}", flush=True)
    path.mkdir(parents=True, exist_ok=True)


def copy_file(source: Path, target: Path) -> None:
    """Copy a file to a target path or directory."""
    print(f"cp {source} {target}", flush=True)
    shutil.copy(source, target)


def copy_glob(pattern: str, target: Path) -> None:
    """Copy files matching a glob pattern to a target directory."""
    for source in ROOT.glob(pattern):
        copy_file(source, target)


def cmake_config(options: argparse.Namespace) -> dict[str, str]:
    """Create CMake configuration from CLI options."""
    config = DEFAULT_CONFIG.copy()
    config["USE_OPENMP"] = options.use_openmp
    config["USE_NVTX"] = options.use_nvtx
    config["PLUGIN_RMM"] = options.plugin_rmm

    if options.log_capi_invocation == "ON":
        config["LOG_CAPI_INVOCATION"] = "ON"
    if options.use_debug == "ON":
        config["CMAKE_BUILD_TYPE"] = "Debug"
    if options.use_cuda == "ON":
        config["USE_CUDA"] = "ON"
        config["USE_NCCL"] = "ON"
        config["USE_DLOPEN_NCCL"] = "OFF"

    return config


def cmake_args(config: dict[str, str]) -> list[str]:
    """Create CMake command line arguments."""
    args = [
        f"-D{k}:BOOL={v}" if v in ("ON", "OFF") else f"-D{k}:STRING={v}"
        for k, v in config.items()
    ]

    if sys.platform != "win32" and shutil.which("ninja"):
        args.append("-GNinja")

    # Set GPU_ARCH_FLAG to override the CUDA architectures.
    if gpu_arch_flag := os.getenv("GPU_ARCH_FLAG"):
        args.append(f"-DCMAKE_CUDA_ARCHITECTURES={gpu_arch_flag}")

    return args


def windows_generators() -> tuple[list[str], ...]:
    """Return CMake generator arguments to try on Windows."""
    return (
        [],  # Let CMake decide.
        ["-G", "Visual Studio 18 2026", "-A", "x64"],
        ["-G", "Visual Studio 17 2022", "-A", "x64"],
        ["-G", "Visual Studio 16 2019", "-A", "x64"],
        ["-G", "Visual Studio 15 2017", "-A", "x64"],
    )


def configure(config_args: list[str], build_dir: Path) -> None:
    """Configure the CMake build."""
    if sys.platform == "win32":
        for generator in windows_generators():
            try:
                run(["cmake", str(ROOT), *config_args, *generator], cwd=build_dir)
                return
            except subprocess.CalledProcessError as err:
                print(
                    f"Failed to build with generator: {shlex.join(generator)}",
                    err,
                    flush=True,
                )
                shutil.rmtree(build_dir)
                mkdir(build_dir)
        raise RuntimeError("None of the supported CMake generators worked.")

    run(["cmake", str(ROOT), *config_args], cwd=build_dir)


def build(config: dict[str, str], build_dir: Path) -> None:
    """Build the native library."""
    if (lib_dir := ROOT / "lib").exists():
        shutil.rmtree(lib_dir)

    configure(cmake_args(config), build_dir)

    build_args = ["cmake", "--build", ".", "--config", "Release"]
    if sys.platform == "linux":
        build_args.extend(["--", "-j", str(os.cpu_count() or 1)])
    elif sys.platform == "win32":
        build_args.extend(
            [
                "--",
                "/m",
                "/nodeReuse:false",
                "/consoleloggerparameters:ShowCommandLine;Verbosity=minimal",
            ]
        )
    run(build_args, cwd=build_dir)


def copy_native_library() -> None:
    """Copy the native library into the JVM package resources."""
    library_name, os_folder = {
        "Windows": ("xgboost4j.dll", "windows"),
        "Darwin": ("libxgboost4j.dylib", "macos"),
        "Linux": ("libxgboost4j.so", "linux"),
        "FreeBSD": ("libxgboost4j.so", "freebsd"),
        "SunOS": ("libxgboost4j.so", "solaris"),
    }[platform.system()]
    arch_folder = {
        "x86_64": "x86_64",  # on Linux & macOS x86_64
        "amd64": "x86_64",  # on Windows x86_64
        "i86pc": "x86_64",  # on Solaris x86_64
        "sun4v": "sparc",  # on Solaris sparc
        "arm64": "aarch64",  # on macOS & Windows ARM 64-bit
        "aarch64": "aarch64",
    }[platform.machine().lower()]

    output_folder = (
        JVM_PACKAGES / "xgboost4j/src/main/resources/lib" / os_folder / arch_folder
    )
    mkdir(output_folder)
    copy_file(ROOT / "lib" / library_name, output_folder)


def copy_test_resources(*, use_cuda: bool) -> None:
    """Copy training data used by JVM package tests."""
    xgboost4j_resources = JVM_PACKAGES / "xgboost4j/src/test/resources"
    mkdir(xgboost4j_resources)
    copy_glob("demo/data/agaricus.*", xgboost4j_resources)

    xgboost4j_spark_resources = JVM_PACKAGES / "xgboost4j-spark/src/test/resources"
    mkdir(xgboost4j_spark_resources)

    regression_dir = ROOT / "demo/data/regression"
    run([sys.executable, "mapfeat.py"], cwd=regression_dir)
    run([sys.executable, "mknfold.py", "machine.txt", "1"], cwd=regression_dir)

    copy_glob("demo/data/regression/machine.txt.t*", xgboost4j_spark_resources)
    copy_glob("demo/data/agaricus.*", xgboost4j_spark_resources)

    if use_cuda:
        xgboost4j_spark_gpu_resources = (
            JVM_PACKAGES / "xgboost4j-spark-gpu/src/test/resources"
        )
        mkdir(xgboost4j_spark_gpu_resources)
        copy_glob("demo/data/veterans_lung_cancer.csv", xgboost4j_spark_gpu_resources)
        copy_file(
            xgboost4j_spark_resources / "rank.train.csv",
            xgboost4j_spark_gpu_resources,
        )


def native_build(options: argparse.Namespace) -> None:
    """Build and copy the native JNI library and its test resources."""
    if sys.platform == "darwin":
        os.environ["JAVA_HOME"] = (
            subprocess.check_output(["/usr/libexec/java_home"]).strip().decode()
        )

    print("building Java wrapper", flush=True)
    build_dir = ROOT / ("build-gpu" if options.use_cuda == "ON" else "build")
    mkdir(build_dir)
    build(cmake_config(options), build_dir)

    print("copying native library", flush=True)
    copy_native_library()

    print("copying train/test files", flush=True)
    copy_test_resources(use_cuda=options.use_cuda == "ON")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--log-capi-invocation", type=str, choices=["ON", "OFF"], default="OFF"
    )
    parser.add_argument("--use-cuda", type=str, choices=["ON", "OFF"], default="OFF")
    parser.add_argument("--use-openmp", type=str, choices=["ON", "OFF"], default="ON")
    parser.add_argument("--use-debug", type=str, choices=["ON", "OFF"], default="OFF")
    parser.add_argument("--use-nvtx", type=str, choices=["ON", "OFF"], default="OFF")
    parser.add_argument("--plugin-rmm", type=str, choices=["ON", "OFF"], default="OFF")
    parsed_args = parser.parse_args()
    native_build(parsed_args)



"""Tracker for XGBoost collective."""

import ctypes
import json
import socket
from enum import IntEnum, unique
from typing import Dict, Optional, Union

from .core import _LIB, _check_call, _deprecate_positional_args, make_jcargs


def get_family(addr: str) -> int:
    """Get network family from address."""
    return socket.getaddrinfo(addr, None)[0][0]


class RabitTracker:
    """Tracker for the collective used in XGBoost, acting as a coordinator between
    workers.

    Parameters
    ----------

    n_workers:

        The total number of workers in the communication group.

    host_ip:

        The IP address of the tracker node. XGBoost can try to guess one by probing with
        sockets. But it's best to explicitly pass an address.

    port:

        The port this tracker should listen to. XGBoost can query an available port from
        the OS, this configuration is useful for restricted network environments.

    sortby:

        How to sort the workers for rank assignment. The default is host, but users can
        set the `DMLC_TASK_ID` via arguments of :py:meth:`~xgboost.collective.init` and
        obtain deterministic rank assignment through sorting by task name. Available
        options are:

          - host
          - task

    timeout :

        Timeout for constructing (bootstrap) and shutting down the communication group,
        doesn't apply to communication when the group is up and running.

        The timeout value should take the time of data loading and pre-processing into
        account, due to potential lazy execution. By default the Tracker doesn't have
        any timeout to avoid pre-mature aborting.

        The :py:meth:`.wait_for` method has a different timeout parameter that can stop
        the tracker even if the tracker is still being used. A value error is raised
        when timeout is reached.

    Examples
    --------

    .. code-block:: python

        from xgboost.tracker import RabitTracker
        from xgboost import collective as coll

        tracker = RabitTracker(host_ip="127.0.0.1", n_workers=2)
        tracker.start()

        with coll.CommunicatorContext(**tracker.worker_args()):
            ret = coll.broadcast("msg", 0)
            assert str(ret) == "msg"

    """

    @unique
    class _SortBy(IntEnum):
        HOST = 0
        TASK = 1

    @_deprecate_positional_args
    def __init__(  # pylint: disable=too-many-arguments
        self,
        n_workers: int,
        host_ip: Optional[str],
        port: int = 0,
        *,
        sortby: str = "host",
        timeout: int = 0,
    ) -> None:

        handle = ctypes.c_void_p()
        if sortby not in ("host", "task"):
            raise ValueError("Expecting either 'host' or 'task' for sortby.")
        if host_ip is not None:
            get_family(host_ip)  # use python socket to stop early for invalid address
        args = make_jcargs(
            host=host_ip,
            n_workers=n_workers,
            port=port,
            dmlc_communicator="rabit",
            sortby=self._SortBy.HOST if sortby == "host" else self._SortBy.TASK,
            timeout=int(timeout),
        )
        _check_call(_LIB.XGTrackerCreate(args, ctypes.byref(handle)))
        self.handle = handle

    def free(self) -> None:
        """Internal function for testing."""
        if hasattr(self, "handle"):
            handle = self.handle
            del self.handle
            _check_call(_LIB.XGTrackerFree(handle))

    def __del__(self) -> None:
        self.free()

    def start(self) -> None:
        """Start the tracker. Once started, the client still need to call the
        :py:meth:`wait_for` method in order to wait for it to finish (think of it as a
        thread).

        """
        _check_call(_LIB.XGTrackerRun(self.handle, make_jcargs()))

    def wait_for(self, timeout: Optional[int] = None) -> None:
        """Wait for the tracker to finish all the work and shutdown. When timeout is
        reached, a value error is raised. By default we don't have timeout since we
        don't know how long it takes for the model to finish training.

        """
        _check_call(_LIB.XGTrackerWaitFor(self.handle, make_jcargs(timeout=timeout)))

    def worker_args(self) -> Dict[str, Union[str, int]]:
        """Get arguments for workers."""
        c_env = ctypes.c_char_p()
        _check_call(_LIB.XGTrackerWorkerArgs(self.handle, ctypes.byref(c_env)))
        assert c_env.value is not None
        env = json.loads(c_env.value)
        return env
